# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-27
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import xml.etree.ElementTree  # nosec: only used for typing
import numpy
import numpy.typing
import pyproj
import typing
import glymur
import logging
import datetime
import shapefile
import shapely.ops
import shapely.prepared
import shapely.geometry
import defusedxml.ElementTree
import scipy.ndimage.morphology
import scipy.interpolate
import re
import PIL.Image
import PIL.ImageDraw
import idf_converter.lib
# import idf_converter.lib.rasterize
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.lib.types import Granule

logger = logging.getLogger(__name__)

DATA_MODEL = 'GRID_YX'

SUPPORTED_VARIABLES = ['{}_TOA_reflectance', '{}_detector_index',
                       '{}_viewing_azimuth', '{}_viewing_zenith',
                       'sun_azimuth', 'sun_zenith',
                       'cloud_mask', 'land_mask',
                       '{}_atmos_reflectance']
BANDNAMES = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07',
             'B08', 'B8A', 'B09', 'B10', 'B11', 'B12']
AngleGridParams = typing.Optional[typing.Tuple[numpy.typing.NDArray, float,
                                               float, float, float]]


class InputPathMissing(IOError):
    """"""
    pass


class MissingLandmaskPath(Exception):
    """"""
    pass


class MissingAtmoslutPath(Exception):
    """"""
    pass


class S2VariableNotSupported(ValueError):
    """Error raised when the S-2 variable is not supported."""
    pass


class S2BandResolutionNotSupported(ValueError):
    """Error raised when the S-2 band resolution is not supported."""
    pass


class XMLNSNotFound(Exception):
    """Error raised when an XML namespace cannot be read from the XML file."""
    def __init__(self, namespace: str) -> None:
        """"""
        self.namespace = namespace


class MissingDetectorMaskPath(Exception):
    """Error raised when the detector mask file cannot be found for a given
    band."""
    def __init__(self, band_name: str) -> None:
        """"""
        self.band_name = band_name


class ZeroQuantificationValue(Exception):
    """Error raised when the quantfication value is 0."""
    pass


class AngleGrid:
    """"""
    col_step: float
    row_step: float
    values: numpy.typing.NDArray


class DetectorFootprint:
    """"""
    lowercorner: typing.List[float]
    uppercorner: typing.List[float]
    polygons: typing.List[numpy.typing.NDArray]
    detector_index: typing.List[int]


class CloudsMask:
    """"""
    lowercorner: typing.Optional[typing.List[float]]
    uppercorner: typing.Optional[typing.List[float]]
    polygons: typing.List[numpy.typing.NDArray]
    cloud_type: typing.List[int]


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    supp_vars = []
    for varname in SUPPORTED_VARIABLES:
        if varname.startswith('{}'):
            supp_vars.append(varname.format('$BAND'))
        else:
            supp_vars.append(varname)
    supp_vars_str = ', '.join(supp_vars)
    bands = ', '.join(BANDNAMES)

    inp = ('    path\tPath of the input file (S2 L1C granule directory)',
           '    variables\tComma-separated list of variable identifiers to'
           ' extract from the input file. Identifiers of supported variables'
           ' are {}. Where $BAND is the band name, ie. in'
           ' {{{}}}.'.format(supp_vars_str, bands),
           '    variable_overrides_VAR\tDefinition of variable attributes'
           ' that are either missing or erroneous in the input file.'
           ' VAR is the identifier of the variable and the value of this'
           ' option is a comma-separated list of key:value. For instance'
           ' here, variable_overrides_B02_TOA_reflectance=valid_max:0.5 allows'
           ' to directly set valid_max of B02 TOA reflectance.',
           '    resolution\tResolution in meters of the variables. Must be'
           ' a multiple of the requested band(s) native resolution.',
           '    rgb_contrast\tIf True, set automatically valid_min/valid_max'
           ' of B02, B03, B04 TOA reflectances in order to optimize RGB'
           ' contrast over ocean.',
           '    land_mask_path\tPath of OSM land polygons shapefile'
           ' (land-polygons-split-4326). Required for RGB contrast strategy'
           ' and land_mask variable.',
           '    atmos_lut_path\tPath of atmospheric reflectance LUT. Required'
           ' for RGB contrast strategy and $BAND_atmos_reflectance variable.',
           '    glymur_nthreads\tSet Glymur number of threads for JPEG2000'
           ' decoding. Default is 1.',
           )
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def downsample(array: numpy.typing.NDArray,
               down_factor: int,
               operation: str = 'mean',
               dtype: typing.Optional[numpy.typing.DTypeLike] = None
               ) -> numpy.typing.NDArray:
    """
    """
    frshp = numpy.array(array.shape, dtype='int32')
    frshp -= numpy.mod(frshp, down_factor)
    dsshp = (frshp / down_factor).astype('int32')
    frsli = (slice(0, frshp[0]), slice(0, frshp[1]))
    rshp = (dsshp[0], down_factor, dsshp[1], down_factor)
    if operation == 'mean':
        # With numpy==1.21.5 and numpy.ma.MaskedArray.mean()
        # we encountered the following error:
        # -> FloatingPointError: underflow encountered in multiply
        # (division by zero ?)
        # As a consequence, we use here sum() and division.
        # (it also makes easier dtype management)
        if numpy.ma.is_masked(array):
            count = array[frsli].reshape(rshp).count(axis=(1, 3))
            array = array[frsli].reshape(rshp).sum(axis=(1, 3), dtype=dtype)
            try:
                array[count != 0] /= count[count != 0]
            except FloatingPointError:
                # Save sign and detect where log scale can be used
                array_sign = numpy.sign(array)
                # count >= 0

                abs_array = numpy.abs(array)

                ok_mask = (abs_array > 0) & (count > 0)
                ok_ind = numpy.where(ok_mask)

                # Perform division in log space
                array[ok_ind] = numpy.log(abs_array[ok_ind])
                array[ok_ind] -= numpy.log(count[ok_ind])

                # Switch back to linear space and restore sign
                array[ok_ind] = array_sign[ok_ind] * numpy.exp(array[ok_ind])
        else:
            array = array[frsli].reshape(rshp).sum(axis=(1, 3), dtype=dtype)
            try:
                array /= down_factor * down_factor
            except FloatingPointError:
                # Save sign and detect where log scale can be used
                array_sign = numpy.sign(array)
                # Assume down_factor > 0

                abs_array = numpy.abs(array)

                ok_mask = (abs_array > 0)
                ok_ind = numpy.where(ok_mask)

                # Perform division in log space
                array[ok_ind] = numpy.log(abs_array[ok_ind])
                array[ok_ind] -= numpy.log(down_factor * down_factor)

                # Switch back to linear space and restore signs
                array[ok_ind] = array_sign[ok_ind] * numpy.exp(array[ok_ind])

    elif operation == 'max':
        array = array[frsli].reshape(rshp).max(axis=(1, 3))

    return array


def rasterize_polygons_from_coords(
                        linear_rings: typing.List[numpy.typing.NDArray],
                        shape: typing.Tuple[int, int],
                        values: typing.Optional[typing.List[int]] = None,
                        dtype: typing.Optional[numpy.typing.DTypeLike] = None
                        ) -> numpy.typing.NDArray:
    """"""
    if values is not None:
        img_type = 'L'
        if dtype is None:
            dtype = 'uint8'
    else:
        img_type = '1'
        if dtype is None:
            dtype = 'bool'
        values = [1] * len(linear_rings)
    img = PIL.Image.new(img_type, shape, 0)
    pencil = PIL.ImageDraw.Draw(img)
    for i, linear_ring in enumerate(linear_rings):
        coords = [(_[0], _[1]) for _ in linear_ring]
        pencil.polygon(coords, fill=values[i], outline=values[i])

    raster = numpy.array(img.getdata(0), dtype=dtype).reshape(shape)
    img.close()

    return raster


def read_atmos_lut(atmos_lut_path: str) -> typing.Dict[str, typing.Any]:
    """
    """
    lut = {}
    with open(atmos_lut_path, 'r') as f:
        lut_lines = f.readlines()
    lut_array = numpy.array([_.rstrip('\n').split(' ') for _ in lut_lines[2:]],
                            dtype='float32')
    lut['sun_zn'] = numpy.unique(lut_array[:, 0])
    lut['view_zn'] = numpy.unique(lut_array[:, 1])
    lut['delta_az'] = numpy.unique(lut_array[:, 2])
    lut_shape = (lut['sun_zn'].size, lut['view_zn'].size,
                 lut['delta_az'].size, lut_array.shape[1] - 3)
    lut['reflectance'] = lut_array[:, 3:].reshape(lut_shape)
    lut['bands'] = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07',
                    'B08', 'B8A', 'B09', 'B10', 'B11', 'B12']
    return lut


class Sentinel2L1CFile:
    """
    """

    NATRES = {'B01': 60, 'B02': 10, 'B03': 10, 'B04': 10,
              'B05': 20, 'B06': 20, 'B07': 20, 'B08': 10,
              'B8A': 20, 'B09': 60, 'B10': 60, 'B11': 20,
              'B12': 20}

    def __init__(self,
                 granule_path: str,
                 glymur_nthreads: typing.Optional[int] = None
                 ) -> None:
        """
        """
        self.granule_path = granule_path
        # Guess compact product naming
        # (introduced in Product Specification Document 14.0)
        granule_name = os.path.basename(granule_path)
        self._compact_naming = granule_name.startswith('L1C_T')
        # Set Glymur number of threads for JPEG2000 decoding
        if glymur_nthreads is not None:
            glymur.set_option('lib.num_threads', glymur_nthreads)
        # Read metadata XML files
        self.metadata: typing.Dict[str, typing.Any] = {}
        self._angle_grids: typing.Dict[str, AngleGrid] = {}
        self._mask_filenames: typing.Dict[str, typing.Dict[str, str]] = {}
        self._read_safe_metadata()
        self._read_granule_metadata()

    def _get_granule_path(self) -> str:
        """
        """
        return self.granule_path

    def _get_safe_path(self) -> str:
        """
        """
        granule_path = self._get_granule_path()
        return os.path.dirname(os.path.dirname(granule_path))

    def _get_granule_metadata_path(self) -> str:
        """
        """
        granule_path = self._get_granule_path()
        if self._compact_naming:
            granmtd_id = 'MTD_TL.xml'
        else:
            granule_id = os.path.basename(granule_path)
            granmtd_id = '{}.xml'.format(granule_id[:-7].replace('MSI', 'MTD'))
        return os.path.join(granule_path, granmtd_id)

    def _get_safe_metadata_path(self) -> str:
        """
        """
        safe_path = self._get_safe_path()
        if self._compact_naming:
            safemtd_id = 'MTD_MSIL1C.xml'
        else:
            safe_id = os.path.basename(safe_path)
            safemtd_id = safe_id.replace('PRD_MSI', 'MTD_SAF')
            safemtd_id = safemtd_id.replace('.SAFE', '.xml')
        return os.path.join(safe_path, safemtd_id)

    def _get_band_path(self, bandname: str) -> str:
        """
        """
        granule_path = self._get_granule_path()
        if self._compact_naming:
            safe_path = self._get_safe_path()
            safe_id = os.path.basename(safe_path)
            _, _, datatake, _, _, proj, _ = safe_id.split('_')
            band_id = '{}_{}_{}.jp2'.format(proj, datatake, bandname)
        else:
            granule_id = os.path.basename(granule_path)
            band_id = '{}_{}.jp2'.format(granule_id[:-7], bandname)
        return os.path.join(granule_path, 'IMG_DATA', band_id)

    def _get_mask_path(self,
                       mask_type: typing.Optional[str],
                       bandname: typing.Optional[str]
                       ) -> typing.Optional[str]:
        """
        """
        fnames = self._mask_filenames
        if mask_type not in fnames:
            return None
        if bandname not in fnames[mask_type]:
            return None
        safe_path = self._get_safe_path()
        fname = fnames[mask_type][bandname]
        mask_path = os.path.join(safe_path, fname)
        return mask_path

    def _read_safe_metadata(self) -> None:
        """
        """
        # Parse SAFE metadata XML
        safemtd_path = self._get_safe_metadata_path()
        tree = defusedxml.ElementTree.parse(safemtd_path)
        root = tree.getroot()
        _n1ns = re.search('{(.+?)}', root.tag)
        if _n1ns is None:
            logger.error('Could not read XML namespace "n1" from input file')
            raise XMLNSNotFound('n1')
        n1ns = _n1ns.group(1)
        xmlns = {'n1': n1ns}
        prod = root.find('n1:General_Info/Product_Info', xmlns)
        query = ['PRODUCT_START_TIME', 'PRODUCT_STOP_TIME', 'PROCESSING_LEVEL',
                 'PRODUCT_TYPE', 'PROCESSING_BASELINE', 'GENERATION_TIME']
        for q in query:
            elem = prod.find(q)
            self.metadata[elem.tag.lower()] = elem.text
        dttk = prod.find('Datatake')
        query = ['SPACECRAFT_NAME', 'DATATAKE_TYPE', 'DATATAKE_SENSING_START',
                 'SENSING_ORBIT_NUMBER', 'SENSING_ORBIT_DIRECTION']
        for q in query:
            elem = dttk.find(q)
            if q == 'SENSING_ORBIT_NUMBER':
                self.metadata[elem.tag.lower()] = int(elem.text)
            else:
                self.metadata[elem.tag.lower()] = elem.text
        image = root.find('n1:General_Info/Product_Image_Characteristics',
                          xmlns)
        for elem in image.findall('Special_Values'):
            if elem.find('SPECIAL_VALUE_TEXT').text == 'NODATA':
                int_value = int(elem.find('SPECIAL_VALUE_INDEX').text)
                self.metadata['nodata_value'] = int_value
            elif elem.find('SPECIAL_VALUE_TEXT').text == 'SATURATED':
                int_value = int(elem.find('SPECIAL_VALUE_INDEX').text)
                self.metadata['saturated_value'] = int_value
        query = ['QUANTIFICATION_VALUE']
        for q in query:
            elem = image.find(q)
            if q == 'QUANTIFICATION_VALUE':
                self.metadata[elem.tag.lower()] = float(elem.text)
            else:
                self.metadata[elem.tag.lower()] = elem.text
        elems = image.findall('Radiometric_Offset_List/RADIO_ADD_OFFSET')
        if len(elems) == 0:
            self.metadata['radio_add_offset'] = None
        else:
            # radio_add_offset to be used in reflectance formula since
            # Product Specification Document 14.9
            self.metadata['radio_add_offset'] = {}
            for elem in elems:
                bandid = int(elem.get('band_id'))
                bandname = self._band_id2name(bandid)
                float_value = float(elem.text)
                self.metadata['radio_add_offset'][bandname] = float_value
        del tree

    def _read_granule_metadata(self) -> None:
        """
        """
        # Parse granule metadata XML
        granmtd_path = self._get_granule_metadata_path()
        tree = defusedxml.ElementTree.parse(granmtd_path)
        root = tree.getroot()
        _n1ns = re.search('{(.+?)}', root.tag)
        if _n1ns is None:
            logger.error('Could not read XML namespace "n1" from input file')
            raise XMLNSNotFound('n1')
        n1ns = _n1ns.group(1)
        xmlns = {'n1': n1ns}
        gnrl = root.find('n1:General_Info', xmlns)
        query = ['TILE_ID', 'SENSING_TIME', 'Archiving_Info/ARCHIVING_CENTRE',
                 'Archiving_Info/ARCHIVING_TIME']
        for q in query:
            elem = gnrl.find(q)
            self.metadata[elem.tag.lower()] = elem.text
        geom = root.find('n1:Geometric_Info', xmlns)
        query = ['Tile_Geocoding/HORIZONTAL_CS_NAME',
                 'Tile_Geocoding/HORIZONTAL_CS_CODE']
        for q in query:
            elem = geom.find(q)
            self.metadata[elem.tag.lower()] = elem.text
        self.metadata['size'] = {}
        for elem in geom.findall('Tile_Geocoding/Size'):
            res = int(elem.get('resolution'))
            nrows = int(elem.find('NROWS').text)
            ncols = int(elem.find('NCOLS').text)
            self.metadata['size'][res] = (nrows, ncols)
        self.metadata['geoposition'] = {}
        for elem in geom.findall('Tile_Geocoding/Geoposition'):
            res = int(elem.get('resolution'))
            ulx = float(elem.find('ULX').text)
            uly = float(elem.find('ULY').text)
            xdim = float(elem.find('XDIM').text)
            ydim = float(elem.find('YDIM').text)
            self.metadata['geoposition'][res] = (ulx, uly, xdim, ydim)
        sun = geom.find('Tile_Angles/Sun_Angles_Grid')
        for anglename in ['Zenith', 'Azimuth']:
            grid_name = 'sun_{}'.format(anglename.lower())
            entry = sun.find(anglename)
            self._angle_grids[grid_name] = self._parse_angle_entry(entry)
        viewings = geom.findall('Tile_Angles/Viewing_Incidence_Angles_Grids')
        for viewing in viewings:
            bandid = int(viewing.get('bandId'))
            bandname = self._band_id2name(bandid)
            detectorid = int(viewing.get('detectorId'))
            for anglename in ['Zenith', 'Azimuth']:
                grid_name = '{}_{:02d}_viewing_{}'.format(bandname, detectorid,
                                                          anglename.lower())
                entry = viewing.find(anglename)
                pentry = self._parse_angle_entry(entry)
                if numpy.isfinite(pentry.values).any():
                    self._angle_grids[grid_name] = pentry
        qual = root.find('n1:Quality_Indicators_Info', xmlns)
        for msk in qual.findall('Pixel_Level_QI/MASK_FILENAME'):
            bandid = msk.get('bandId')
            mtype = msk.get('type')
            fname = msk.text
            masked_bandname = None
            if bandid is not None:
                masked_bandname = self._band_id2name(int(bandid))
            if mtype not in self._mask_filenames:
                self._mask_filenames[mtype] = {}
            self._mask_filenames[mtype][masked_bandname] = fname
        del tree

    # note: static method
    def _parse_angle_entry(self,
                           entry: xml.etree.ElementTree.Element
                           ) -> AngleGrid:
        """
        """
        col_step_node = entry.find('COL_STEP')
        if col_step_node is None or col_step_node.text is None:
            raise
        col_step = float(col_step_node.text)

        row_step_node = entry.find('ROW_STEP')
        if row_step_node is None or row_step_node.text is None:
            raise
        row_step = float(row_step_node.text)

        values = entry.findall('Values_List/VALUES')
        if values is None or 0 == len(values) or values[0].text is None:
            raise
        shape = (len(values), len(values[0].text.split(' ')))
        grid = numpy.zeros(shape, dtype='float32')
        for line, val_node in enumerate(values):
            if val_node is None or val_node.text is None:
                raise
            grid[line, :] = numpy.array([float(e)
                                         for e in val_node.text.split(' ')])

        result = AngleGrid()
        result.col_step = col_step
        result.row_step = row_step
        result.values = grid
        return result

    def _read_detfoo_gml(self, detfoo_gml_path: str
                         ) -> DetectorFootprint:
        """
        """
        detfoo = DetectorFootprint()
        tree = defusedxml.ElementTree.parse(detfoo_gml_path)
        root = tree.getroot()
        xmlns = {'eop': 'http://www.opengis.net/eop/2.0',
                 'gml': 'http://www.opengis.net/gml/3.2'}
        # Envelope
        lowc = root.find('gml:boundedBy/gml:Envelope/gml:lowerCorner', xmlns)
        detfoo.lowercorner = [float(crd) for crd in lowc.text.split(' ')]
        upc = root.find('gml:boundedBy/gml:Envelope/gml:upperCorner', xmlns)
        detfoo.uppercorner = [float(crd) for crd in upc.text.split(' ')]
        # Detector polygons
        numpy_guess = -1
        xpath = 'eop:extentOf/gml:Polygon/gml:exterior/gml:LinearRing/gml:posList'  # noqa: E501
        polygons = []
        detector_index = []
        for feature in root.findall('eop:maskMembers/eop:MaskFeature', xmlns):
            # mask_type_node = feature.find('eop:maskType', xmlns)
            # if ((mask_type_node is None)
            #     or ('DETECTOR_FOOTPRINT' != mask_type_node.text)):
            #     continue

            _id = feature.get('{{{}}}id'.format(xmlns['gml']))
            detector_index.append(int(_id.split('-')[2]))

            coords_node = feature.find(xpath, xmlns)
            coords_count = int(coords_node.get('srsDimension'))
            _coords = [float(_) for _ in coords_node.text.split(' ')]
            _coords = numpy.array(_coords).reshape((numpy_guess, coords_count))
            polygons.append(_coords)
        detfoo.polygons = polygons
        detfoo.detector_index = detector_index
        return detfoo

    def _read_clouds_gml(self, clouds_gml_path: str
                         ) -> CloudsMask:
        """
        """
        clouds = CloudsMask()
        tree = defusedxml.ElementTree.parse(clouds_gml_path)
        root = tree.getroot()
        xmlns = {'eop': 'http://www.opengis.net/eop/2.0',
                 'gml': 'http://www.opengis.net/gml/3.2'}
        # Envelope
        lowc = root.find('gml:boundedBy/gml:Envelope/gml:lowerCorner', xmlns)
        if lowc is None:
            clouds.lowercorner = None
        else:
            clouds.lowercorner = [float(crd) for crd in lowc.text.split(' ')]
        upc = root.find('gml:boundedBy/gml:Envelope/gml:upperCorner', xmlns)
        if upc is None:
            clouds.uppercorner = None
        else:
            clouds.uppercorner = [float(crd) for crd in upc.text.split(' ')]
        # cloud polygons
        numpy_guess = -1
        xpath = 'eop:extentOf/gml:Polygon/gml:exterior/gml:LinearRing/gml:posList'  # noqa: E501
        polygons = []
        cloud_type = []  # 0=OPAQUE / 1=CIRRUS
        mask_types = ['OPAQUE', 'CIRRUS']
        for feature in root.findall('eop:maskMembers/eop:MaskFeature', xmlns):
            mask_type_node = feature.find('eop:maskType', xmlns)
            # if (mask_type_node is None) or ('OPAQUE' != mask_type_node.text):
            #     continue
            mask_type = mask_type_node.text
            cloud_type.append(mask_types.index(mask_type))

            coords_node = feature.find(xpath, xmlns)
            coords_count = int(coords_node.get('srsDimension'))
            _coords = [float(_) for _ in coords_node.text.split(' ')]
            _coords = numpy.array(_coords).reshape((numpy_guess, coords_count))
            polygons.append(_coords)
        clouds.polygons = polygons
        clouds.cloud_type = cloud_type
        return clouds

    def _get_angle_grid(self,
                        angle_type: str,
                        bandname: typing.Optional[str] = None,
                        detector: typing.Optional[int] = None
                        ) -> AngleGridParams:
        """
        """
        angle_name = angle_type
        if bandname is not None and detector is not None:
            angle_name = '{}_{:02d}_{}'.format(bandname, detector, angle_name)
        if angle_name not in self._angle_grids:
            return None
        grid = self._angle_grids[angle_name]
        values = grid.values.copy()
        ny, nx = values.shape
        ulx, uly, _dx, _dy = self.get_metadata()['geoposition'][10]
        dx = numpy.abs(grid.col_step) * numpy.sign(_dx)
        dy = numpy.abs(grid.row_step) * numpy.sign(_dy)
        xctr0 = ulx
        yctr0 = uly
        isfin = numpy.isfinite(values)
        if 'viewing' in angle_name and isfin.any():
            # Viewing angles are given over each detector area with some NaN
            # around. Here we extrapolate around detector borders in order to
            # avoid NaN inside detector area after future interpolation.
            # First we fit a polynom with valid pixels
            max_deg = 3
            deg = [min([numpy.sum(isfin.sum(axis=1) > 0) - 1, max_deg]),
                   min([numpy.sum(isfin.sum(axis=0) > 0) - 1, max_deg])]
            indfin = numpy.where(isfin)
            vdr = numpy.polynomial.polynomial.polyvander2d(indfin[0],
                                                           indfin[1],
                                                           deg)
            coeff, rs, rk, s = numpy.linalg.lstsq(vdr, values[indfin],
                                                  rcond=-1)
            coeff = coeff.reshape((deg[0] + 1, deg[1] + 1))
            # Then we apply this polynom to the neighbors of valid pixels
            struct = numpy.ones((3, 3), dtype='bool')
            dilate = scipy.ndimage.morphology.binary_dilation
            dilisfin = dilate(isfin, structure=struct, border_value=0)
            ind2fill = numpy.where(~isfin & dilisfin)
            values[ind2fill] = numpy.polynomial.polynomial.polyval2d(
                ind2fill[0], ind2fill[1], coeff)
        return values, dy, dx, yctr0, xctr0

    def get_bandnames(self) -> typing.List[str]:
        """
        """
        return list(self.NATRES.keys())

    def get_native_resolution(self, bandname: str) -> int:
        """
        """
        return self.NATRES[bandname]

    def _band_id2name(self, bandid: int) -> str:
        """
        """
        bandnames = self.get_bandnames()
        return bandnames[bandid]

    def get_metadata(self) -> typing.Dict[str, typing.Any]:
        """
        """
        return self.metadata

    def _downsampling_factor(self, target_resolution: int,
                             native_resolution: int) -> int:
        """
        """
        tres = int(target_resolution)
        nres = int(native_resolution)
        if tres <= 0:
            raise Exception('resolution must be > 0')
        if tres % nres != 0:
            raise Exception('resolution must be multiple of native one')
        return int(tres / nres)

    def check_resolution(self, resolution: int, bandname: str) -> bool:
        """
        """
        tres = int(resolution)
        if tres <= 0:
            return False
        nres = self.get_native_resolution(bandname)
        if tres % nres != 0:
            return False
        return True

    def get_shape(self,
                  resolution: int = 60
                  ) -> typing.Tuple[int, int]:
        """
        """
        downfac = self._downsampling_factor(resolution, 10)
        natny, natnx = self.get_metadata()['size'][10]
        ny = int(numpy.floor(natny / downfac))
        nx = int(numpy.floor(natnx / downfac))
        return (ny, nx)

    def _get_grid_def(self,
                      resolution: int
                      ) -> typing.Tuple[int, int, int, int]:
        """
        """
        ulx, uly, xdim, ydim = self.get_metadata()['geoposition'][10]
        dx = resolution * numpy.sign(xdim)
        dy = resolution * numpy.sign(ydim)
        return ulx, uly, dx, dy

    def get_xy(self,
               resolution: int = 60
               ) -> typing.Tuple[numpy.typing.NDArray, numpy.typing.NDArray]:
        """
        """
        ulx, uly, dx, dy = self._get_grid_def(resolution)
        ny, nx = self.get_shape(resolution)
        y = numpy.arange(ny, dtype='float32') * dy + dy / 2. + uly
        x = numpy.arange(nx, dtype='float32') * dx + dx / 2. + ulx
        return x, y

    def get_lonlat(self,
                   resolution: int = 60
                   ) -> typing.Tuple[numpy.typing.NDArray,
                                     numpy.typing.NDArray]:
        """
        """
        epsg = self.get_metadata()['horizontal_cs_code']
        proj = pyproj.Proj(init=epsg)
        x, y = self.get_xy(resolution)
        nx = x.size
        ny = y.size
        x = numpy.tile(x[numpy.newaxis, :], (ny, 1))
        y = numpy.tile(y[:, numpy.newaxis], (1, nx))
        lon, lat = proj(x, y, inverse=True)
        return lon, lat

    def get_detector_index(self,
                           bandname: str,
                           resolution: int = 60
                           ) -> numpy.typing.NDArray:
        """
        """
        mask_path = self._get_mask_path('MSK_DETFOO', bandname)
        if mask_path is None:
            raise MissingDetectorMaskPath(bandname)

        if mask_path.endswith('.jp2'):
            natres = self.get_native_resolution(bandname)
            downfac = self._downsampling_factor(resolution, natres)
            # Raster mask in JP2 format
            # Since Product Specification Document 14.8
            handler = glymur.Jp2k(mask_path)
            detfoo = handler[:]  # uint8
            if downfac > 1:
                i0 = int((downfac - 1) / 2.)
                detfoo = detfoo[i0::downfac, i0::downfac]
                ny, nx = self.get_shape(resolution)
                detfoo = detfoo[:ny, :nx]
        else:
            shape = self.get_shape(resolution)
            ulx, uly, dx, dy = self._get_grid_def(resolution)
            # Polygons in GML format, to be rasterized
            # Before Product Specification Document 14.8
            detfoo_gml = self._read_detfoo_gml(mask_path)
            linear_rings = []
            for poly in detfoo_gml.polygons:
                ring = numpy.stack(((poly[:, 0] - ulx) / dx,
                                    (poly[:, 1] - uly) / dy), axis=1)
                linear_rings.append(ring)
            values = detfoo_gml.detector_index
            detfoo = rasterize_polygons_from_coords(linear_rings, shape,
                                                    values)
            # import matplotlib.pyplot as plt
            # extent = [0, shape[1], 0, shape[0]]
            # plt.imshow(detfoo, origin='lower', interpolation='nearest',
            #            extent=extent)
            # for poly in linear_rings:
            #     plt.plot(poly[:, 0], poly[:, 1], '+-', color='k')
            # plt.show()
            # import pdb ; pdb.set_trace()
        return detfoo

    def _get_sun_angle(self,
                       angle_type: str,
                       resolution: int = 60
                       ) -> numpy.typing.NDArray:
        """
        """
        x, y = self.get_xy(resolution)
        _tmp = self._get_angle_grid(angle_type)
        if _tmp is None:
            return None
        val, dy, dx, yc0, xc0 = _tmp
        # Bilinear interpolation
        # Initial grid may contain NaN
        fy = (y[:, numpy.newaxis] - yc0) / dy
        iy = numpy.floor(fy).astype('int32')
        iy = numpy.clip(iy, 0, val.shape[0] - 2, out=iy)
        fx = (x[numpy.newaxis, :] - xc0) / dx
        ix = numpy.floor(fx).astype('int32')
        ix = numpy.clip(ix, 0, val.shape[1] - 2, out=ix)
        wy2 = fy - iy
        wy1 = 1. - wy2
        wx2 = fx - ix
        wx1 = 1. - wx2
        angle = wx1 * (wy1 * val[iy, ix] + wy2 * val[iy + 1, ix]) + \
            wx2 * (wy1 * val[iy, ix + 1] + wy2 * val[iy + 1, ix + 1])
        angle = numpy.ma.masked_invalid(angle, copy=False)
        return angle

    def get_sun_azimuth(self,
                        resolution: int = 60
                        ) -> numpy.typing.NDArray:
        """
        """
        return self._get_sun_angle('sun_azimuth', resolution)

    def get_sun_zenith(self,
                       resolution: int = 60
                       ) -> numpy.typing.NDArray:
        """
        """
        return self._get_sun_angle('sun_zenith', resolution)

    def _get_viewing_angle(self,
                           angle_type: str,
                           bandname: str,
                           resolution: int = 60
                           ) -> numpy.typing.NDArray:
        """
        """
        x, y = self.get_xy(resolution)
        det_ind = self.get_detector_index(bandname, resolution)
        hist, _ = numpy.histogram(det_ind, numpy.arange(1, 14))
        angle = numpy.ma.masked_all(det_ind.shape, dtype='float32')
        for idet in range(1, 13):
            if hist[idet - 1] == 0:
                continue
            ind = numpy.where(det_ind == idet)
            _tmp = self._get_angle_grid(angle_type, bandname, idet)
            if _tmp is None:
                continue
            val, dy, dx, yc0, xc0 = _tmp
            # Bilinear interpolation
            # Initial grid may contain NaN
            fy = (y[ind[0]] - yc0) / dy
            iy = numpy.floor(fy).astype('int32')
            iy = numpy.clip(iy, 0, val.shape[0] - 2, out=iy)
            fx = (x[ind[1]] - xc0) / dx
            ix = numpy.floor(fx).astype('int32')
            ix = numpy.clip(ix, 0, val.shape[1] - 2, out=ix)
            wy2 = fy - iy
            wy1 = 1. - wy2
            wx2 = fx - ix
            wx1 = 1. - wx2
            _val = wx1 * (wy1 * val[iy, ix] + wy2 * val[iy + 1, ix]) + \
                wx2 * (wy1 * val[iy, ix + 1] + wy2 * val[iy + 1, ix + 1])
            _val = numpy.ma.masked_invalid(_val, copy=False)
            angle[ind] = _val
        return angle

    def get_viewing_azimuth(self,
                            bandname: str,
                            resolution: int = 60
                            ) -> numpy.typing.NDArray:
        """
        """
        return self._get_viewing_angle('viewing_azimuth', bandname, resolution)

    def get_viewing_zenith(self,
                           bandname: str,
                           resolution: int = 60
                           ) -> numpy.typing.NDArray:
        """
        """
        return self._get_viewing_angle('viewing_zenith', bandname, resolution)

    def get_digital_number(self,
                           bandname: str,
                           resolution: int = 60
                           ) -> numpy.typing.NDArray:
        """
        """
        natres = self.get_native_resolution(bandname)
        downfac = self._downsampling_factor(resolution, natres)
        band_path = self._get_band_path(bandname)
        handler = glymur.Jp2k(band_path)
        dnum = numpy.ma.masked_equal(handler[:], 0)  # uint16
        if downfac > 1:
            dnum = downsample(dnum, downfac, dtype='float32')
            dnum = dnum.round().astype('uint16')
        return dnum

    def get_toa_reflectance(self,
                            bandname: str,
                            resolution: int = 60
                            ) -> numpy.typing.NDArray:
        """
        """
        natres = self.get_native_resolution(bandname)
        downfac = self._downsampling_factor(resolution, natres)
        metadata = self.get_metadata()
        toar = self.get_digital_number(bandname, natres).astype('float32')
        offset = metadata['radio_add_offset']
        if offset is not None:
            toar += offset[bandname]
        try:
            toar /= metadata['quantification_value']
        except FloatingPointError:
            # Save sign and detect where log scale can be used
            toar_sign = numpy.sign(toar)
            denom_sign = numpy.sign(metadata['quantification_value'])

            abs_toar = numpy.abs(toar)
            abs_denom = numpy.abs(metadata['quantification_value'])

            if 0 == abs_denom:
                raise ZeroQuantificationValue()

            ok_mask = (abs_toar > 0)
            ok_ind = numpy.where(ok_mask)

            # Perform division in log space
            toar[ok_ind] = numpy.log(abs_toar[ok_ind])
            toar[ok_ind] -= numpy.log(abs_denom)

            # Switch back to linear space and restore signs
            toar[ok_ind] = (toar_sign[ok_ind] * denom_sign *
                            numpy.exp(toar[ok_ind]))

        if downfac > 1:
            toar = downsample(toar, downfac, dtype='float32')
        return toar

    def get_cloud_mask(self, resolution: int = 60
                       ) -> typing.Optional[numpy.typing.NDArray]:
        """
        """
        mask_path1 = self._get_mask_path('MSK_CLASSI', None)
        mask_path2 = self._get_mask_path('MSK_CLOUDS', None)
        if mask_path1 is not None:
            tres = int(resolution)
            if tres <= 0:
                raise Exception('resolution must be > 0')
            if tres % 60 == 0:
                upfac = 1
                downfac = self._downsampling_factor(tres, 60)
            elif tres % 30 == 0:
                upfac = 2
                downfac = self._downsampling_factor(tres, 30)
            elif tres % 20 == 0:
                upfac = 3
                downfac = self._downsampling_factor(tres, 20)
            elif tres % 10 == 0:
                upfac = 6
                downfac = self._downsampling_factor(tres, 10)
            else:
                raise Exception('resolution must be multiple of native one')
            # Raster mask in JP2 format
            # Since Product Specification Document 14.8
            # Warning: MSK_CLASSI is given at 60m resolution
            # band1 = Opaque clouds
            # band2 = Cirrus clouds
            # band3 = Snow and Ice areas
            handler = glymur.Jp2k(mask_path1)
            cloud_mask = handler[:, :, :2].sum(axis=-1) != 0
            if upfac != 1:
                iy = numpy.repeat(numpy.arange(cloud_mask.shape[0]), upfac)
                ix = numpy.repeat(numpy.arange(cloud_mask.shape[1]), upfac)
                cloud_mask = cloud_mask[iy[:, numpy.newaxis],
                                        ix[numpy.newaxis, :]]
            if downfac != 1:
                cloud_mask = downsample(cloud_mask, downfac, operation='max')
        elif mask_path2 is not None:
            shape = self.get_shape(resolution)
            ulx, uly, dx, dy = self._get_grid_def(resolution)
            # Polygons in GML format, to be rasterized
            # Before Product Specification Document 14.8
            clouds_gml = self._read_clouds_gml(mask_path2)
            linear_rings = []
            for poly in clouds_gml.polygons:
                ring = numpy.stack(((poly[:, 0] - ulx) / dx,
                                    (poly[:, 1] - uly) / dy), axis=1)
                linear_rings.append(ring)
            cloud_mask = rasterize_polygons_from_coords(linear_rings, shape)
        else:
            cloud_mask = None
        return cloud_mask

    def get_land_mask(self,
                      land_mask_path: str,
                      resolution: int = 60
                      ) -> numpy.typing.NDArray:
        """
        """
        ulx, uly, dx, dy = self._get_grid_def(resolution)
        ny, nx = self.get_shape(resolution)
        # Set x/y polygon surrounding the data
        ul_corner = (ulx, uly)
        ur_corner = (ulx + nx * dx, uly)
        br_corner = (ulx + nx * dx, uly + ny * dy)
        bl_corner = (ulx, uly + ny * dy)
        _bbox_polygon = shapely.geometry.Polygon([ul_corner, ur_corner,
                                                  br_corner, bl_corner])
        # Landmask is provided with lon/lat polygons so the bbox coordinates
        # must be reprojected
        wgs84 = pyproj.CRS('EPSG:4326')
        epsg_code = self.get_metadata()['horizontal_cs_code']
        utm = pyproj.CRS(epsg_code)
        utm_to_wgs84 = pyproj.Transformer.from_crs(utm, wgs84, always_xy=True)
        _bbox_polygon = shapely.ops.transform(utm_to_wgs84.transform,
                                              _bbox_polygon)
        # Prepare the polygon to optimize intersection test in the upcoming
        # loop
        bbox_polygon = shapely.prepared.prep(_bbox_polygon)
        # Later on the land polygons will be used in UTM coordinates, so
        # prepare the projection
        wgs84_to_utm = pyproj.Transformer.from_crs(wgs84, utm, always_xy=True)
        # Only keep landmask shapes that intersect the granule bbox
        related_shapes = []
        f = shapefile.Reader(land_mask_path)
        if shapefile.__version__ >= '2.2.0':
            bbox = _bbox_polygon.bounds
            shapes = f.iterShapes(bbox=bbox)
        else:
            shapes = f.iterShapes()
        for shape in shapes:
            shapely_shape = shapely.geometry.shape(shape)
            if not bbox_polygon.intersects(shapely_shape):
                continue
            intersection_shape = _bbox_polygon.intersection(shapely_shape)
            utm_shape = shapely.ops.transform(wgs84_to_utm.transform,
                                              intersection_shape)
            related_shapes.append(utm_shape)
        f.close()
        # print('{} landmask polygons found'.format(len(related_shapes)))
        # import pdb ; pdb.set_trace()

        # Rasterize landmask shapes
        linear_rings = []
        for shape in related_shapes:
            if isinstance(shape, shapely.geometry.Polygon):
                x, y = numpy.array(shape.exterior.coords.xy)
                ring = numpy.stack(((x - ulx) / dx,
                                    (y - uly) / dy), axis=1)
                linear_rings.append(ring)
            elif isinstance(shape, shapely.geometry.MultiPolygon):
                for polygon in shape.geoms:
                    x, y = numpy.array(polygon.exterior.coords.xy)
                    ring = numpy.stack(((x - ulx) / dx,
                                        (y - uly) / dy), axis=1)
                    linear_rings.append(ring)
            # else:
            #     print(type(shape))
        if len(linear_rings) == 0:
            land_mask = numpy.zeros((ny, nx), dtype='bool')
        else:
            land_mask = rasterize_polygons_from_coords(linear_rings, (ny, nx))
        return land_mask

    # def get_atmos_reflectance(self, atmos_lut_path, bandname, resolution=60):
    #     """
    #     """
    #     # Get detector index and angles
    #     det_ind = self.get_detector_index(bandname, resolution)
    #     sun_az = self.get_sun_azimuth(resolution)
    #     sun_zn = self.get_sun_zenith(resolution)
    #     view_az = self.get_viewing_azimuth(bandname, resolution)
    #     view_zn = self.get_viewing_zenith(bandname, resolution)
    #     # Read LUT and set interpolator
    #     atmos_lut = read_atmos_lut(atmos_lut_path)
    #     lut_points = (atmos_lut['sun_zn'], atmos_lut['view_zn'],
    #                   atmos_lut['delta_az'])
    #     lut_iband = atmos_lut['bands'].index(bandname)
    #     lut_values = atmos_lut['reflectance'][:, :, :, lut_iband]
    #     interpolator = scipy.interpolate.RegularGridInterpolator
    #     lut_func = interpolator(lut_points, lut_values, method='linear',
    #                             bounds_error=False, fill_value=None)
    #     # Apply interpolator per detector
    #     hist, _ = numpy.histogram(det_ind, numpy.arange(1, 14))
    #     atmosr = numpy.ma.masked_all(det_ind.shape, dtype='float32')
    #     for idet in range(1, 13):
    #         if hist[idet - 1] == 0:
    #             continue
    #         ind = numpy.where(det_ind == idet)
    #         saz, szn = sun_az[ind], sun_zn[ind]
    #         vaz, vzn = view_az[ind], view_zn[ind]
    #         delta_az = numpy.abs(numpy.mod(saz - vaz + 180., 360.) - 180.)
    #         points = numpy.array((szn, vzn, delta_az)).transpose()
    #         atmosr[ind] = lut_func(points)
    #     return atmosr

    def get_atmos_reflectance(self,
                              atmos_lut_path: str,
                              bandname: str,
                              resolution: int = 60
                              ) -> numpy.typing.NDArray:
        """
        """
        # Get angles
        sun_az = self.get_sun_azimuth(resolution)
        sun_zn = self.get_sun_zenith(resolution)
        view_az = self.get_viewing_azimuth(bandname, resolution)
        view_zn = self.get_viewing_zenith(bandname, resolution)
        # Read LUT and set interpolator
        atmos_lut = read_atmos_lut(atmos_lut_path)
        lut_points = (atmos_lut['sun_zn'], atmos_lut['view_zn'],
                      atmos_lut['delta_az'])
        lut_iband = atmos_lut['bands'].index(bandname)
        lut_values = atmos_lut['reflectance'][:, :, :, lut_iband]
        interpolator = scipy.interpolate.RegularGridInterpolator
        lut_func = interpolator(lut_points, lut_values, method='linear',
                                bounds_error=False, fill_value=None)
        # Apply interpolator
        atmosr = numpy.ma.masked_all(sun_az.shape, dtype='float32')
        ind = numpy.where(~numpy.ma.getmaskarray(sun_az) &
                          ~numpy.ma.getmaskarray(sun_zn) &
                          ~numpy.ma.getmaskarray(view_az) &
                          ~numpy.ma.getmaskarray(view_zn))
        if ind[0].size == 0:
            return atmosr
        saz, szn = sun_az[ind], sun_zn[ind]
        vaz, vzn = view_az[ind], view_zn[ind]
        delta_az = numpy.abs(numpy.mod(saz - vaz + 180., 360.) - 180.)
        points = numpy.array((szn, vzn, delta_az)).transpose()
        atmosr[ind] = lut_func(points)
        return atmosr


def dilate_mask(values: numpy.typing.NDArray,
                dilation_length: float,
                x_spacing: float,
                y_spacing: float
                ) -> numpy.typing.NDArray:
    """"""
    center_px = 1
    branches_px: typing.List[int]
    branches_px = [numpy.round(dilation_length / sp).astype('int')
                   for sp in (y_spacing, x_spacing)]
    dilation_shape = [2 * branch_px + center_px for branch_px in branches_px]
    dilation_kernel = numpy.ones(dilation_shape, dtype='bool')
    dilate = scipy.ndimage.morphology.binary_dilation
    result = dilate(values, structure=dilation_kernel)
    return result


def set_rgb_contrast(granule: Granule, sen2l1c: Sentinel2L1CFile,
                     resol: int, land_mask_path: str, atmos_lut_path: str
                     ) -> None:
    """"""
    rgb_id = ['B04_TOA_reflectance', 'B03_TOA_reflectance',
              'B02_TOA_reflectance']

    exists = []
    for var_id in rgb_id:
        vmin = granule.vars[var_id]['valid_min']
        vmax = granule.vars[var_id]['valid_max']
        exists.append(vmin is not None and vmax is not None)
    if all(exists):
        return

    # Set resolution for contrast computation
    target_ctst_resol = 80
    if resol >= target_ctst_resol:
        ctst_resol = resol
        downfac = 1
    else:
        downfac = int(numpy.round(target_ctst_resol / float(resol)))
        ctst_resol = resol * downfac

    # Get reflectance data
    rgb_array = []
    for var_id in rgb_id:
        array = granule.vars[var_id]['array']
        if downfac != 1:
            array = downsample(array, downfac, dtype='float32')
        rgb_array.append(array)

    # Set mask
    # Reflectance masks
    mask = numpy.zeros(rgb_array[0].shape, dtype='bool')
    for array in rgb_array:
        mask |= numpy.ma.getmaskarray(array)
    # Land mask
    if 'land_mask' in granule.vars:
        land_mask = granule.vars['land_mask']['array']
        if downfac != 1:
            land_mask = downsample(land_mask, downfac, operation='max')
    else:
        land_mask = sen2l1c.get_land_mask(land_mask_path, ctst_resol)
    dilation_length = 500.  # meters
    mask |= dilate_mask(land_mask, dilation_length, ctst_resol, ctst_resol)
    # Cloud mask
    if 'cloud_mask' in granule.vars:
        cloud_mask = granule.vars['cloud_mask']['array']
        if downfac != 1:
            cloud_mask = downsample(cloud_mask, downfac, operation='max')
    else:
        cloud_mask = sen2l1c.get_cloud_mask(ctst_resol)
    if cloud_mask is not None:
        dilation_length = 500.  # meters
        mask |= dilate_mask(cloud_mask, dilation_length, ctst_resol,
                            ctst_resol)
    if numpy.all(mask):
        return

    # vmin
    # vmin from data
    _vmin_data = []
    for array in rgb_array:
        pmask = (numpy.ma.getmaskarray(array) | mask)
        _vmin_data.append(numpy.percentile(numpy.ma.getdata(array[~pmask]),
                                           0.5))
    # vmin from atmos lut
    _vmin_atm = []
    for var_id in rgb_id:
        bandname = var_id[:3]
        atm_var_id = '{}_atmos_reflectance'.format(bandname)
        if atm_var_id in granule.vars:
            array = granule.vars[atm_var_id]['array']
        else:
            array = sen2l1c.get_atmos_reflectance(atmos_lut_path, bandname,
                                                  ctst_resol)
        _vmin_atm.append(array.mean())
    # final vmin
    shifts = [_v - _v_atm for _v, _v_atm in zip(_vmin_data, _vmin_atm)]
    shift = min(shifts)
    _vmin = [_v_atm + shift for _v_atm in _vmin_atm]

    # vmax
    slope_threshold = -40.
    _vmax_data = []
    bins = numpy.linspace(0, .5, num=51)
    for array in rgb_array:
        hist, edges = numpy.histogram(array[~mask], bins=bins)
        dbin = edges[1] - edges[0]
        hist = hist.astype('float') / hist.sum() / dbin
        slope = (hist[1:] - hist[:-1]) / dbin
        slope_ctrs = edges[1:-1]
        ind = numpy.where(slope <= slope_threshold)[0]
        if ind.size != 0:
            ind = ind.max()
            dy = slope[ind + 1] - slope[ind]
            dx = slope_ctrs[ind + 1] - slope_ctrs[ind]
            _vm = slope_ctrs[ind] + (slope_threshold - slope[ind]) / dy * dx
            _vmax_data.append(_vm)
        else:
            pmask = (numpy.ma.getmaskarray(array) | mask)
            _vmax_data.append(numpy.percentile(numpy.ma.getdata(array[~pmask]),
                                               99.5))
    vrans = [vma - vmi for vma, vmi in zip(_vmax_data, _vmin)]
    vran = max(vrans)
    _vmax = [vmi + vran for vmi in _vmin]

    # Fill
    for ibnd, var_id in enumerate(rgb_id):
        vmin = granule.vars[var_id]['valid_min']
        if vmin is None:
            granule.vars[var_id]['valid_min'] = _vmin[ibnd]
        vmax = granule.vars[var_id]['valid_max']
        if vmax is None:
            granule.vars[var_id]['valid_max'] = _vmax[ibnd]


def compute_minmax(vars_id: typing.List[str],
                   granule: Granule
                   ) -> None:
    """"""
    for var_id in vars_id:
        if var_id.endswith('_TOA_reflectance'):
            vmin = granule.vars[var_id]['valid_min']
            vmax = granule.vars[var_id]['valid_max']
            if vmin is None or vmax is None:
                values = numpy.ma.compressed(granule.vars[var_id]['array'])
                if vmin is None:
                    _vmin = numpy.percentile(values, 0.1)
                    granule.vars[var_id]['valid_min'] = _vmin
                if vmax is None:
                    _vmax = numpy.percentile(values, 99.9)
                    granule.vars[var_id]['valid_max'] = _vmax
        elif var_id.endswith('_detector_index'):
            vmin = granule.vars[var_id]['valid_min']
            vmax = granule.vars[var_id]['valid_max']
            if vmin is None:
                granule.vars[var_id]['valid_min'] = 0
            if vmax is None:
                granule.vars[var_id]['valid_max'] = 254
        elif var_id.endswith('_azimuth') or var_id.endswith('_zenith'):
            vmin = granule.vars[var_id]['valid_min']
            vmax = granule.vars[var_id]['valid_max']
            if vmin is None:
                _vmin = granule.vars[var_id]['array'].min()
                granule.vars[var_id]['valid_min'] = _vmin
            if vmax is None:
                _vmax = granule.vars[var_id]['array'].max()
                granule.vars[var_id]['valid_max'] = _vmax
        elif var_id in ['cloud_mask', 'land_mask']:
            vmin = granule.vars[var_id]['valid_min']
            vmax = granule.vars[var_id]['valid_max']
            if vmin is None:
                granule.vars[var_id]['valid_min'] = 0
            if vmax is None:
                granule.vars[var_id]['valid_max'] = 254
        elif var_id.endswith('_atmos_reflectance'):
            vmin = granule.vars[var_id]['valid_min']
            vmax = granule.vars[var_id]['valid_max']
            if vmin is None:
                _vmin = granule.vars[var_id]['array'].min()
                granule.vars[var_id]['valid_min'] = _vmin
            if vmax is None:
                _vmax = granule.vars[var_id]['array'].max()
                granule.vars[var_id]['valid_max'] = _vmax


def init_variables(vars_id: typing.List[str]
                   ) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
    """"""
    variables: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
    for var_id in vars_id:
        if var_id.endswith('_TOA_reflectance'):
            bandname = var_id[:3]
            long_name = '{} top-of-atmosphere reflectance'.format(bandname)
            units = None  # dimensionless ?
        elif var_id.endswith('_detector_index'):
            bandname = var_id[:3]
            long_name = '{} detector index'.format(bandname)
            units = None
        elif var_id.endswith('_viewing_azimuth'):
            bandname = var_id[:3]
            long_name = '{} viewing azimuth angle'.format(bandname)
            units = 'degree'
        elif var_id.endswith('_viewing_zenith'):
            bandname = var_id[:3]
            long_name = '{} viewing zenith angle'.format(bandname)
            units = 'degree'
        elif var_id == 'sun_azimuth':
            long_name = 'sun azimuth angle'
            units = 'degree'
        elif var_id == 'sun_zenith':
            long_name = 'sun zenith angle'
            units = 'degree'
        elif var_id == 'cloud_mask':
            long_name = 'cloud mask'
            units = None
        elif var_id == 'land_mask':
            long_name = 'land mask'
            units = None
        elif var_id.endswith('_atmos_reflectance'):
            bandname = var_id[:3]
            long_name = '{} atmospheric reflectance'.format(bandname)
            units = None
        else:
            raise S2VariableNotSupported()
        variables[var_id] = {'name': var_id,
                             'long_name': long_name,
                             'valid_min': None,
                             'valid_max': None,
                             'options': {}}
        if units is not None:
            variables[var_id]['units'] = units
    return variables


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    # Get inputs
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    _vars_id = input_opts.get('variables', None)
    if _vars_id is None:
        # Default
        vars_id = ['B02_TOA_reflectance',
                   'B03_TOA_reflectance',
                   'B04_TOA_reflectance']
    else:
        vars_id = _vars_id.split(',')

    _resol = input_opts.get('resolution', None)
    if _resol is None:
        # Default (will work for any band)
        resol = 60
    else:
        resol = int(_resol)

    _rgb_ctst = input_opts.get('rgb_contrast', 'False')
    rgb_contrast = _rgb_ctst in ['True', 'true']

    # OSM: land-polygons-split-4326/land_polygons.shp
    land_mask_path = input_opts.get('land_mask_path', None)

    atmos_lut_path = input_opts.get('atmos_lut_path', None)

    glymur_nthreads = input_opts.get('glymur_nthreads', None)
    if glymur_nthreads is not None:
        glymur_nthreads = int(glymur_nthreads)

    # Check inputs
    # Check vars_id
    supp_vars_all = []
    for varname in SUPPORTED_VARIABLES:
        if varname.startswith('{}'):
            for bandname in BANDNAMES:
                supp_vars_all.append(varname.format(bandname))
        else:
            supp_vars_all.append(varname)
    for var_id in vars_id:
        if var_id not in supp_vars_all:
            raise S2VariableNotSupported()
    # Disable rgb_contrast if RGB bands are not all requested
    if rgb_contrast:
        has_rgb = []
        for bandname in ['B02', 'B03', 'B04']:
            _var_id = '{}_TOA_reflectance'.format(bandname)
            has_rgb.append(_var_id in vars_id)
        rgb_contrast = all(has_rgb)
    # Check land mask path
    asking_land = 'land_mask' in vars_id
    if rgb_contrast or asking_land:
        if land_mask_path is None:
            raise MissingLandmaskPath()
    # Check atmos lut path
    asking_atmos = any([v.endswith('_atmos_reflectance') for v in vars_id])
    if rgb_contrast or asking_atmos:
        if atmos_lut_path is None:
            raise MissingAtmoslutPath()
    # Init Sentinel2L1CFile
    sen2l1c = Sentinel2L1CFile(input_path, glymur_nthreads=glymur_nthreads)
    # Check resolution
    for var_id in vars_id:
        _bandname = var_id[:3]
        if _bandname in BANDNAMES:
            if not sen2l1c.check_resolution(resol, _bandname):
                raise S2BandResolutionNotSupported()

    # Init granule
    variables = init_variables(vars_id)

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables
    idf_converter.lib.apply_var_overrides(input_opts, granule)

    # Read variables
    array: typing.Optional[typing.Any]
    for var_id in vars_id:
        if var_id.endswith('_TOA_reflectance'):
            bandname = var_id[:3]
            array = sen2l1c.get_toa_reflectance(bandname, resol)
        elif var_id.endswith('_detector_index'):
            bandname = var_id[:3]
            array = sen2l1c.get_detector_index(bandname, resol)
        elif var_id.endswith('_viewing_azimuth'):
            bandname = var_id[:3]
            array = sen2l1c.get_viewing_azimuth(bandname, resol)
        elif var_id.endswith('_viewing_zenith'):
            bandname = var_id[:3]
            array = sen2l1c.get_viewing_zenith(bandname, resol)
        elif var_id == 'sun_azimuth':
            array = sen2l1c.get_sun_azimuth(resol)
        elif var_id == 'sun_zenith':
            array = sen2l1c.get_sun_zenith(resol)
        elif var_id == 'cloud_mask':
            array = sen2l1c.get_cloud_mask(resol)
        elif var_id == 'land_mask':
            array = sen2l1c.get_land_mask(land_mask_path, resol)
        elif var_id.endswith('_atmos_reflectance'):
            bandname = var_id[:3]
            array = sen2l1c.get_atmos_reflectance(atmos_lut_path, bandname,
                                                  resol)
        granule.vars[var_id]['array'] = array

    # Set valid_min / valid_max
    if rgb_contrast:
        set_rgb_contrast(granule, sen2l1c, resol, land_mask_path,
                         atmos_lut_path)

    compute_minmax(vars_id, granule)

    # Geolocation
    ny, nx = sen2l1c.get_shape(resol)
    granule.dims['y'] = ny
    granule.dims['x'] = nx
    x_values, y_values = sen2l1c.get_xy(resol)
    granule.vars['y'] = {'array': y_values,
                         'datatype': y_values.dtype,
                         'options': {}}
    granule.vars['x'] = {'array': x_values,
                         'datatype': x_values.dtype,
                         'options': {}}
    output_opts['gcp_spacing'] = int(25000 / resol)
    metadata = sen2l1c.get_metadata()
    epsg = metadata['horizontal_cs_code']
    input_opts['projection'] = epsg

    # Metadata
    s2_granule_root = os.path.dirname(input_path)
    s2_granule_count = 0
    for bname in os.listdir(s2_granule_root):
        path = os.path.join(s2_granule_root, bname)
        if os.path.isdir(path):
            s2_granule_count += 1
    if s2_granule_count > 1:
        # Multitile product (old products at beginning of S2A mission start)
        s2_granule_name = os.path.basename(input_path)
        idf_granule_name = '{}-{}m'.format(s2_granule_name, resol)
    else:
        # Nomical case: 1 granule (tile) per S2 product
        s2_safe_root = os.path.dirname(s2_granule_root)
        s2_safe_name = os.path.splitext(os.path.basename(s2_safe_root))[0]
        idf_granule_name = '{}-{}m'.format(s2_safe_name, resol)

    time_fmt = '%Y-%m-%dT%H:%M:%S.%fZ'  # eg. 2022-03-25T11:21:09.024Z
    start_time = metadata['product_start_time']
    start_dt = datetime.datetime.strptime(start_time, time_fmt)
    stop_time = metadata['product_stop_time']
    stop_dt = datetime.datetime.strptime(stop_time, time_fmt)
    platform = metadata['spacecraft_name']

    granule.meta['idf_granule_id'] = idf_granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = resol
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt
    granule.meta['institution'] = 'ESA'
    granule.meta['platform'] = platform
    granule.meta['platform_type'] = 'leo satellite'
    granule.meta['sensor'] = 'MSI'
    granule.meta['sensor_type'] = 'multi-spectral imager'

    # Add transforms
    transforms = []

    # Artificially increase the range on a linear scale, thus allowing
    # users to desaturate the rendering of this channel in SEAScope
    # _max_linear = _min_linear + 2 * (_max_linear - _min_linear)
    if rgb_contrast:
        rgb_vars = ['B02_TOA_reflectance', 'B03_TOA_reflectance',
                    'B04_TOA_reflectance']
        transforms.append(('stretch_extent', {'targets': rgb_vars,
                                              'stretch_factor': 2.0,
                                              'change_vmin': False,
                                              'change_vmax': True}))

    output_opts['__export'] = vars_id

    yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import numpy
import typing
import logging
import scipy.signal
import numpy.typing
import scipy.interpolate
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)

FractionalGrid = typing.Tuple[numpy.typing.NDArray, numpy.typing.NDArray]
Contour = typing.Tuple[numpy.typing.NDArray, numpy.typing.NDArray]


def gen_fractional_grid(param: typing.Dict[str, typing.Any]) -> FractionalGrid:
    """" Generate fractional grid from xml file parameters """
    # retrieve parameters
    elem_n = param['elem_n']  # Across-track
    line_n = param['line_n']  # Along-track
    # Track offset nadir
    toffset_n = param['trackOffset_n']
    # Track offset tie point
    toffset_tx = param['trackOffset_tx']
    # Start offset nadir
    soffset_n = param['startOffset_n']
    # Start offset tie point
    soffset_tx = param['startOffset_tx']
    # Spatial resolution nadir
    res_n = param['res_n']
    # Spatial resolution tie points
    res_tx = param['res_tx']
    # Interpolating tie point fractional grid
    dj_in = numpy.zeros((elem_n), dtype='float')
    di_in = numpy.zeros((line_n), dtype='float')

    # Build fractional grid
    x_tx = toffset_tx * res_tx
    y_tx = soffset_tx * res_n
    start_index = 0
    for j in range(elem_n):
        x_in = (toffset_n - j) * res_n
        dj_in[j] = (x_tx - x_in) / res_tx

    for j in range(start_index, line_n + start_index):
        y_in = (soffset_n + j) * res_n
        di_in[j-start_index] = (y_in - y_tx) / res_n

    return di_in, dj_in


def filter_sst(sst: numpy.typing.NDArray,
               bt8: numpy.typing.NDArray,
               kernel_size: typing.Tuple[int, int]
               ) -> numpy.typing.NDArray:
    filter_kernel: numpy.typing.ArrayLike
    filter_kernel = numpy.ones(kernel_size) / (kernel_size[0] * kernel_size[1])
    sst_diff: numpy.typing.ArrayLike = sst - bt8
    diff_mean = scipy.signal.convolve2d(sst_diff, filter_kernel, mode='same',
                                        boundary='fill', fillvalue=numpy.nan)
    sst_filtered: numpy.typing.NDArray = diff_mean + bt8
    sst_filtered[0, :] = sst[0, :]
    sst_filtered[-1, :] = sst[-1, :]

    return sst_filtered


def get_alongtrack_contour(values: numpy.typing.NDArray,
                           offset: int
                           ) -> Contour:
    """Expects along-track to be the first axis"""
    alongtrack_size = numpy.shape(values)[0]
    idx_i = []
    idx_j = []
    values_mask = numpy.ma.getmaskarray(values)
    for i in range(0, alongtrack_size):
        valid_values = numpy.where(values_mask[i] == False)  # noqa
        if 0 >= numpy.size(valid_values):
            continue
        first: numpy.typing.ScalarType = numpy.min(valid_values[0])
        idx_i.append(i)
        idx_j.append(first)
        for o in range(0, offset):
            idx_i.append(i)
            idx_j.append(first + 1 + o)

        last: numpy.typing.ScalarType = numpy.max(valid_values[0])
        if first != last:
            idx_i.append(i)
            idx_j.append(last)
            for o in range(0, offset):
                idx_i.append(i)
                idx_j.append(min(0, last - (1 + o)))
    result = (numpy.array(idx_i, dtype=int), numpy.array(idx_j, dtype=int))
    return result


def interp_to_grid(coeffs_n2: numpy.typing.NDArray,
                   tcwv: numpy.typing.NDArray,
                   path_tn: numpy.typing.NDArray,
                   di_in: numpy.typing.NDArray,
                   dj_in: numpy.typing.NDArray,
                   elem_n: int,
                   line_n: int
                   ) -> numpy.typing.NDArray:
    """Interpolate TCWV and Pathlength to fractional grid"""
    # Interpolate TCWV
    nrow_metx = numpy.arange((numpy.shape(tcwv)[1]))
    ncol_metx = numpy.arange((numpy.shape(tcwv)[0]))
    interpolator = scipy.interpolate.RegularGridInterpolator
    points = (ncol_metx, nrow_metx)

    # Interpolate TCWV
    func = interpolator(points, tcwv, method='linear',
                        bounds_error=False, fill_value=None)
    _di, _dj = numpy.meshgrid(di_in, dj_in)
    _tcwv = func((_di.ravel(), _dj.ravel()))
    tcwv_in = _tcwv.reshape(elem_n, line_n)  # tcwv at full resolution

    # Interpolate Pathlength
    func = interpolator(points, path_tn, method='nearest',
                        bounds_error=False, fill_value=None)
    _path = func((_di.ravel(), _dj.ravel()))
    path_in = _path.reshape(elem_n, line_n)  # path at full resolution

    # - Interpolate SST coefficients to fractional grid
    coeffs_n2_single = coeffs_n2[:, 0, :, :]
    n2_in = numpy.zeros((elem_n, line_n, 3))
    nrow_n2 = numpy.arange((numpy.shape(coeffs_n2_single)[2]))
    ncol_n2 = numpy.arange((numpy.shape(coeffs_n2_single)[1]))
    points_n2 = (ncol_n2, nrow_n2)
    for index in range(3):
        func = interpolator(points_n2, coeffs_n2_single[index, :, :],
                            method='linear', bounds_error=False,
                            fill_value=None)
        _n2 = func((tcwv_in.ravel(), path_in.ravel()))
        n2_in[:, :, index] = _n2.reshape((elem_n, line_n))
    return n2_in


def n2_sst_from_slstr(input_opts: InputOptions,
                      output_opts: OutputOptions,
                      granule: Granule,
                      *args: typing.Iterable[typing.Any],
                      input_data: typing.Dict[str, typing.Any]
                      ) -> FormatterResult:
    """"""
    view = 'n'  # Nadir view
    tcwv = input_data['tcwv']
    sat_zen_n = input_data['sat_zen_n']
    param = input_data['param']
    coeffs_n2 = input_data['coeffs_n2']
    n11_id = input_data['n11_id']
    n12_id = input_data['n12_id']

    # Copy data to make sure there will be no side effect
    n11 = numpy.ma.copy(granule.vars[n11_id]['array'])
    n12 = numpy.ma.copy(granule.vars[n12_id]['array'])

    # Look for the first column which has a non-masked value
    # Transpose so that the first index corresponds to the first valid
    # longitude (and not the first valid latitude)
    contour_n11 = get_alongtrack_contour(n11, 0)
    n11.mask[contour_n11] = True

    # Convert zenith angles into pathlength
    path_tn = 1.0 / numpy.cos(numpy.deg2rad(sat_zen_n))

    # - Generate fractional grid
    elem_n = param[f'elem_{view}']
    line_n = param[f'line_{view}']
    di_in, dj_in = gen_fractional_grid(param)

    # - Interpolate TCWV and Pathlength to fractional grid
    n2_in = interp_to_grid(coeffs_n2, tcwv, path_tn,
                           di_in, dj_in, elem_n, line_n)

    alpha = n2_in[:, :, 0].transpose()
    beta = n2_in[:, :, 1].transpose()
    gamma = n2_in[:, :, 2].transpose()
    band = (alpha + beta * n12[:, :] + gamma * n11[:, :])
    kernel_size = (3, 3)
    band = filter_sst(band, n11, kernel_size)

    # Remove borders + n to remove pixels affected by the kernel_size kernel
    # n = int((kernel_size[0] - 1) / 2)
    contour_byte = get_alongtrack_contour(n11, 1)
    band[contour_byte] = numpy.nan

    # Delete data copies
    del n11
    del n12

    granule.vars['sst']['array'] = band

    return input_opts, output_opts, granule



# vim: ts=4:sts=4:sw=4
#
# @date 2020-01-07
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
import dateutil.relativedelta
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger(__name__)

DATA_MODEL = idf_converter.readers.netcdf_grid_latlon.DATA_MODEL


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


class MissingYearStart(Exception):
    """Error raised when the "year_start" input option has not been specified
    by the user."""
    pass


class MissingYearEnd(Exception):
    """Error raised when the "year_end" input option has not been specified by
    the user."""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\t'.join(out))


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    # Read variables
    f_handler = netCDF4.Dataset(input_path, 'r')
    _time = f_handler.variables['time'][:]
    _time_units = f_handler.variables['time'].units
    f_handler.close()

    _year_start = input_opts.get('year_start', None)
    if _year_start is None:
        raise MissingYearStart()
    year_start = int(_year_start)

    _year_end = input_opts.get('year_end', None)
    if _year_end is None:
        raise MissingYearEnd()
    year_end = int(_year_end)

    input_opts['lon_variable'] = 'lon'
    input_opts['lat_variable'] = 'lat'
    input_opts['depth_variable'] = 'depth'
    input_opts['depth_dimension'] = 'depth'
    input_opts['time_coverage_relative_start'] = 'month_beginning'
    input_opts['time_coverage_relative_end'] = 'month_end'

    # 2009
    if not _time_units.startswith('months since'):
        input_opts['time_variable'] = 'time'

    generic_reader = idf_converter.readers.netcdf_grid_latlon.read_data
    for year in range(year_start, 1 + year_end):
        if _time_units.startswith('months since'):
            # 2013 / 2018

            # Numpy internal computations for modulo can lead to underflow
            # errors because numbers have a small order of magnitude, so
            # scale operands up by a large factor, perform the modulo and scale
            # the result down the the same factor
            UFLOW_FIX = 10**6
            scaled_month = numpy.mod(_time * UFLOW_FIX, 12 * UFLOW_FIX)
            month = numpy.floor(scaled_month / UFLOW_FIX).astype('int')
            ref_dt = datetime.datetime(year, 1, 1)
            dt = ref_dt + dateutil.relativedelta.relativedelta(months=month)
            input_opts['indicative_times'] = numpy.array([dt, ])
        else:
            # 2009
            pattern = _time_units.replace('2008', f'{year}')
            pattern = pattern.replace('0000', f'{year}')
            input_opts['time_units'] = pattern.format(year)

        generic_results = generic_reader(input_opts, output_opts)
        for input_opts, output_opts, granule, transforms in generic_results:
            yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

import os
import numpy
import typing
import logging
import netCDF4

logger = logging.getLogger(__name__)

DIMNAMES = ['lat', 'lon', 'y', 'x', 'row', 'cell']
DIMGCPNAMES = ['{}_gcp'.format(dimname) for dimname in DIMNAMES]
VARGCPNAMES = ['lat_gcp', 'lon_gcp']
VARINDEXGCPNAMES = ['index_{}_gcp'.format(dimname) for dimname in DIMNAMES]


class MultiTimeGranuleError(ValueError):
    """Error raised when the input IDF file contains more than one time value.
    Note: downscaling is not supported for time-dependent data model."""
    pass


class UnknownDimension(ValueError):
    """Error raised when a dimension of the input file is neither a temporal,
    a spatial or a GCP dimension."""
    def __init__(self, dim_name: str) -> None:
        """"""
        self.dim_name = dim_name


class MaskedValuesFound(ValueError):
    """Error raised when masked values are found in the data array before it is
    assigned to a variable in the output file. At this point the data array
    should contain fill/missing values instead of masked ones."""
    pass


class DownscaleFactorNotSupported(ValueError):
    """Error raised when the downscale factor is so big that it would not be
    possible to define a standard data type for storing the count of valid
    data (i.e. risk of overflowing uint64 data type)."""
    def __init__(self, factor: int) -> None:
        """"""
        self.factor = factor


class IncompatibleVariableShape(ValueError):
    """Error raised when the shape of the input variable does not count exactly
    three axes, which is the shape expected for geophysical variables in IDF
    files. It means that the downscaling method does not support extra
    dimensions such as depth or (k, phi)."""
    def __init__(self, name: str, shape: typing.Tuple[int]) -> None:
        """"""
        self.var_name = name
        self.var_shape = shape


class InvalidGCPIndexShape(ValueError):
    """Error raised when a GCP index variable from the input file has more than
    one dimension, which should never happen."""
    def __init__(self, name: str, shape: typing.Tuple[int]) -> None:
        """"""
        self.var_name = name
        self.var_shape = shape


class MaxGCPIndexDoesNotMatchGCPDimension(ValueError):
    """Error raised if the max value stored in the GCP index variable for an
    axis will not match the size of the dimension associated with this axis."""
    def __init__(self, name: str, dim_name: str, max_value: int,
                 dim_size: int) -> None:
        """"""
        self.var_name = name
        self.dim_name = dim_name
        self.dim_size = dim_size
        self.max_value = max_value


def downscale(idf_path: str,
              output_dir: typing.Optional[str] = None,
              min_factor: typing.Optional[int] = None,
              max_factor: typing.Optional[int] = None) -> None:
    """
    """
    if output_dir is None:
        output_dir = os.path.dirname(idf_path)

    os.makedirs(output_dir, exist_ok=True)

    input_idf = netCDF4.Dataset(idf_path, 'r')
    input_factor = input_idf.idf_subsampling_factor
    if min_factor is None:
        min_factor = input_factor + 1
    if max_factor is None:
        max_factors = []
        # compute how many times each dimension can be subsampled
        for input_dim_name, input_dim in input_idf.dimensions.items():
            if input_dim_name in DIMGCPNAMES:
                float_factor = numpy.log(len(input_dim) - 1) / numpy.log(2.)
                _factor = numpy.ceil(float_factor).astype('int')
                max_factors.append(input_factor + _factor)
        max_factor = max(max_factors)

    for factor in range(min_factor, max_factor + 1):
        downfactor = 2 ** (factor - input_factor)
        fname = '{}_idf_{:02d}.nc'.format(input_idf.idf_granule_id, factor)
        output_path = os.path.join(output_dir, fname)
        output_idf = netCDF4.Dataset(output_path, 'w', format='NETCDF4')

        # Dimensions
        for input_dim_name, input_dim in input_idf.dimensions.items():
            if 'time' == input_dim_name:
                if len(input_dim) != 1:
                    raise MultiTimeGranuleError
                if input_dim.isunlimited():
                    output_dim = output_idf.createDimension(input_dim_name,
                                                            None)
                else:
                    output_dim = output_idf.createDimension(input_dim_name,
                                                            len(input_dim))
            elif input_dim_name in DIMNAMES:
                _dimlen = len(input_dim) / float(downfactor)
                dimlen = numpy.ceil(_dimlen).astype('int32')
                output_dim = output_idf.createDimension(input_dim_name, dimlen)  # noqa
            elif input_dim_name in DIMGCPNAMES:
                _dimlen = (len(input_dim) - 1) / float(downfactor)
                dimlen = numpy.ceil(_dimlen).astype('int32') + 1
                output_dim = output_idf.createDimension(input_dim_name, dimlen)  # noqa
            else:
                raise UnknownDimension(input_dim_name)

        # Variables
        for input_var_name, input_var in input_idf.variables.items():
            _filters = input_var.filters()
            zlib = _filters.get('zlib', True)
            complevel = _filters.get('complevel', 4)
            shuffle = _filters.get('shuffle', True)
            fletcher32 = _filters.get('fletcher32', False)
            contiguous = ('contiguous' == input_var.chunking())
            endian = input_var.endian()
            lsd = input_var.quantization()

            # Cannot use the same chunking as the original for the downscaled
            # IDF files as the dimensions have different sizes
            chunksizes = None

            if '_FillValue' in input_var.ncattrs():
                fvalue = input_var._FillValue
                output_var = output_idf.createVariable(
                                                input_var_name,
                                                input_var.datatype,
                                                input_var.dimensions,
                                                zlib=zlib,
                                                complevel=complevel,
                                                shuffle=shuffle,
                                                fletcher32=fletcher32,
                                                contiguous=contiguous,
                                                chunksizes=chunksizes,
                                                least_significant_digit=lsd,
                                                endian=endian,
                                                fill_value=fvalue)
            else:
                output_var = output_idf.createVariable(
                                                input_var_name,
                                                input_var.datatype,
                                                input_var.dimensions,
                                                zlib=zlib,
                                                complevel=complevel,
                                                shuffle=shuffle,
                                                fletcher32=fletcher32,
                                                contiguous=contiguous,
                                                chunksizes=chunksizes,
                                                least_significant_digit=lsd,
                                                endian=endian)

            for attrname in input_var.ncattrs():
                if '_FillValue' == attrname:
                    continue
                output_var.setncattr(attrname, input_var.getncattr(attrname))
            if input_var_name == 'time':
                output_var[:] = input_var[:]
            elif (input_var_name == 'crs') and 'y_gcp' in input_idf.dimensions:
                output_var[:] = input_var[:]
            elif input_var_name in VARGCPNAMES:
                subind = []
                for idim in range(len(input_var.shape)):
                    _subind = numpy.arange(0, input_var.shape[idim],
                                           downfactor)
                    if numpy.mod(input_var.shape[idim] - 1, downfactor) != 0:
                        _subind = numpy.append(_subind,
                                               input_var.shape[idim] - 1)
                    _shape = numpy.ones(len(input_var.shape)).astype('int32')
                    _shape[idim] = -1
                    subind.append(_subind.reshape(_shape))
                output_var[:] = input_var[:][tuple(subind)]
            elif input_var_name in VARINDEXGCPNAMES:
                if len(input_var.shape) > 1:
                    raise InvalidGCPIndexShape(input_var_name, input_var.shape)
                subind = numpy.arange(0, input_var.shape[0], downfactor)
                if numpy.mod(input_var.shape[0] - 1, downfactor) != 0:
                    subind = numpy.append(subind, input_var.shape[0] - 1)
                # ceil may be needed for last index (not necessarily modulo
                # downfactor)
                output = numpy.ceil(input_var[subind] / float(downfactor))
                output = output.astype(output_var.datatype)
                dimname = input_var.dimensions[0].split('_')[0]
                dimsize = len(output_idf.dimensions[dimname])
                if output[-1] != dimsize:
                    raise MaxGCPIndexDoesNotMatchGCPDimension(input_var_name,
                                                              dimname,
                                                              output[-1],
                                                              dimsize)

                output_var[:] = output
            else:
                if len(input_var.shape) != 3:
                    raise IncompatibleVariableShape(input_var_name,
                                                    input_var.shape)
                # We let netCDF4 auto masking when reading input
                # but for the rest we take care of it.
                # Some strange behaviors:
                # - auto scaling input: it appears float64 are used internally
                # even when scale_factor and add_offset are float32
                # - auto masking output (with auto scaling off): masked values
                #   are not replaced by _FillValue
                input_var.set_auto_mask(True)
                input_var.set_auto_scale(False)
                output_var.set_auto_maskandscale(False)
                # Datatype down_type is used for downsampling (ie averaging)
                # Datatype cnt_type is used for counting valid data
                down_dtype = 'float32'
                if downfactor < 2 ** 4:
                    cnt_dtype = 'uint8'
                elif downfactor < 2 ** 8:
                    cnt_dtype = 'uint16'
                elif downfactor < 2 ** 16:
                    cnt_dtype = 'uint32'
                elif downfactor < 2 ** 32:
                    cnt_dtype = 'uint64'
                else:
                    raise DownscaleFactorNotSupported(downfactor)
                # Read input
                dshape = output_var.shape[1:]
                fshape = (dshape[0] * downfactor, dshape[1] * downfactor)
                rshape = (dshape[0], downfactor, dshape[1], downfactor)
                output = numpy.ma.MaskedArray(numpy.zeros(fshape,
                                                          dtype=down_dtype),
                                              mask=numpy.ones(fshape,
                                                              dtype='bool'))
                input_slice = (slice(None, input_var.shape[1]),
                               slice(None, input_var.shape[2]))
                output[input_slice] = input_var[0, :, :]
                # Average in blocks of size downfactor x downfactor
                # (we avoid .mean() function which does not seem to respect
                # dtype and uses float64)
                output = output.reshape(rshape)
                output_mask = numpy.ma.getmaskarray(output)
                cnt_valid = (~output_mask).sum(axis=(1, 3), dtype=cnt_dtype)
                output = output.sum(axis=(1, 3), dtype=down_dtype)
                # (output should be masked where cnt_valid=0)
                try:
                    output /= cnt_valid
                except FloatingPointError:
                    # Save sign and detect where log scale can be used
                    sign = numpy.sign(output)
                    abs_output = numpy.abs(output)
                    ok_mask = (cnt_valid > 0) & (abs_output > 0)
                    ok_ind = numpy.where(ok_mask)

                    # Perform division in log space
                    output[ok_ind] = numpy.log(abs_output[ok_ind])
                    output[ok_ind] -= numpy.log(cnt_valid[ok_ind])

                    # Switch back to linear space and restore sign
                    output[ok_ind] = sign[ok_ind] * numpy.exp(output[ok_ind])

                if numpy.issubdtype(output_var.dtype, numpy.integer):
                    output = output.round().astype(output_var.dtype)
                else:
                    output = output.astype(output_var.dtype)

                # Update mask in order to mask where too few valid points
                exist_mask = numpy.zeros(fshape, dtype='bool')
                exist_mask[:input_var.shape[1], :input_var.shape[2]] = True
                cnt_exist = exist_mask.reshape(rshape).sum(axis=(1, 3),
                                                           dtype=cnt_dtype)
                del exist_mask
                output.mask = cnt_valid < (cnt_exist / 2.)
                del cnt_valid, cnt_exist
                # Put into netCDF4 variable
                if 'missing_value' in output_var.ncattrs():
                    output = output.filled(output_var.missing_value)
                elif '_FillValue' in output_var.ncattrs():
                    output = output.filled(output_var._FillValue)
                else:
                    if numpy.ma.is_masked(output):
                        raise MaskedValuesFound()
                output_var[0, :, :] = output

        # Global attributes
        for attrname in input_idf.ncattrs():
            if 'idf_subsampling_factor' == attrname:
                attrvalue = numpy.int32(factor)
            elif 'idf_spatial_resolution' == attrname:
                attrvalue = input_idf.getncattr(attrname) * downfactor
            else:
                attrvalue = input_idf.getncattr(attrname)
            output_idf.setncattr(attrname, attrvalue)
        output_idf.close()
    input_idf.close()



# vim: ts=4:sts=4:sw=4
#
# @date 2020-02-07
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.


""""""

import numpy
import typing
import logging
import scipy.ndimage.filters
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)

# Scale factor used to circumvent underflow errors
UFLOW_FIX = 10**6


def s1_radvel_exp(input_opts: InputOptions,
                  output_opts: OutputOptions,
                  granule: Granule,
                  *args: typing.Iterable[typing.Any],
                  targets: typing.Tuple[str],
                  outputs: typing.Tuple[str],
                  incidence_angle: str,
                  sweep_angle: str) -> FormatterResult:
    """"""
    _sweep = granule.vars[sweep_angle]['array']
    _incidence = granule.vars[incidence_angle]['array']
    sinus_incidence = numpy.sin(numpy.deg2rad(_incidence)
                                ).astype(numpy.float64)

    for input_var, output_var in zip(targets, outputs):
        _input = granule.vars[input_var]['array']

        # Descalloping
        corr = (- 20 * _sweep * (abs(_sweep) <= 0.1)
                - 8 * (_sweep + 0.15) * ((_sweep > 0.1) & (_sweep <= 0.54))
                + 50 * (_sweep - 0.65) * (_sweep > 0.54)
                - 8 * (_sweep - 0.15) * ((_sweep < - 0.1) & (_sweep >= - 0.54))
                + 50 * (_sweep + 0.65) * (_sweep < - 0.54))
        corr = corr.astype(numpy.float64) / 106

        # smooth
        smoothed = scipy.ndimage.median_filter(_input + corr, size=(3, 3))

        if output_var not in granule.vars.keys():
            granule.vars[output_var] = {'name': output_var,
                                        'options': {}}

        # Underflow may happen in numpy internal computations when dividing
        # smoothed by sinus_incidence.
        # In this specific case there is not much of a difference between the
        # order of magnitude of the operands, the problem is simply that values
        # are too close to zero, so multiply the numerator and the denominator
        # by the same large number prior to computing the division.
        radvel = (UFLOW_FIX * smoothed) / (UFLOW_FIX * sinus_incidence)
        granule.vars[output_var]['array'] = radvel

        if 'valid_min' not in granule.vars[output_var].keys():
            vmin: numpy.typing.ScalarType
            vmin = numpy.nanmin(granule.vars[output_var]['array'])
            granule.vars[output_var]['valid_min'] = vmin

        if 'valid_max' not in granule.vars[output_var].keys():
            vmax: numpy.typing.ScalarType
            vmax = numpy.nanmax(granule.vars[output_var]['array'])
            granule.vars[output_var]['valid_max'] = vmax

    return (input_opts, output_opts, granule)


def radial_velocity_sign(input_opts: InputOptions,
                         output_opts: OutputOptions,
                         granule: Granule,
                         *args: typing.Iterable[typing.Any],
                         targets: typing.Tuple[str],
                         update_extrema: bool,
                         heading: str) -> FormatterResult:
    """"""
    # Note: assume heading provided in degrees, clockwise from north
    mean_heading = granule.vars[heading]['array'].mean()

    mean_heading_trigo = numpy.deg2rad(90.0 - mean_heading)
    if 0 < numpy.sin(mean_heading_trigo):
        for target in targets:
            _values = granule.vars[target]['array']
            granule.vars[target]['array'] = -1.0 * _values
            if update_extrema is True:
                vmin = granule.vars[target]['valid_min']
                vmax = granule.vars[target]['valid_max']
                granule.vars[target]['valid_min'] = -1.0 * vmax
                granule.vars[target]['valid_max'] = -1.0 * vmin

    return (input_opts, output_opts, granule)



# vim: ts=4:sts=4:sw=4
#
# @date 2020-09-30
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.


TIMEFMT = '%Y-%m-%dT%H:%M:%S.%fZ'

SUPPORTED_MODELS = {
    'grid_latlon': ('time', 'lat', 'lon', True, True),
    'grid_yx': ('time', 'y', 'x', False, True),
    'swath': ('time', 'row', 'cell', False, False),
    'time_dependent': ('time', False, False)}

MANDATORY_ATTRIBUTES = (
    'idf_version',
    'idf_granule_id',
    'idf_subsampling_factor',
    'idf_spatial_resolution',
    'idf_spatial_resolution_units',
    'time_coverage_start',
    'time_coverage_end')

SUPPORTED_ATTRIBUTES = (
    'title',
    'institution',
    'comment',
    'id',
    'product_version',
    'geospatial_lat_max',
    'geospatial_lat_min',
    'geospatial_lon_max',
    'geospatial_lon_min',
    'creator_email')

CF_ATTRIBUTES = (
    'Conventions',
    'summary',
    'references',
    'institution_abbreviation',
    'history',
    'naming_authority',
    'processing_software',
    'uuid',
    'date_created',
    'date_modified',
    'keywords',
    'keywords_vocabulary',
    'publisher_name',
    'publisher_url')

OPTIONAL_ATTRIBUTES = (
    'license',
    'netcdf_version_id',
    'file_quality_level',
    'spatial_resolution',
    'time_coverage_resolution',
    'geospatial_vertical_min',
    'geospatial_vertical_max',
    'geospatial_vertical_units',
    'geospatial_vertical_positive',
    'nominal_latitude',
    'nominal_longitude',
    'source',
    'source_version',
    'wmo_id',
    'buoy_network',
    'station_name',
    'sea_floor_depth_below_sea_level',
    'site_elevation',
    'platform_type',
    'band',
    'sensor_description',
    'sensor_manufacturer',
    'sensor_serial_number',
    'sensor_install_date',
    'sensor_height',
    'sensor_sampling_period',
    'sensor_sampling_rate',
    'sensor_calibration_date',
    'sensor_history',
    'sensor_type',
    'Metadata_Conventions',
    'metadata_link',
    'standard_name_vocabulary',
    'acknowledgement',
    'creator_name',
    'creator_url',
    'publisher_email',
    'processing_level',
    'cdm_data_type')

SPECIAL_ATTRIBUTES = (
    'platform',
    'sensor',
    'station_id',
    'forecast_range',
    'run_time',
    'forecast_type',
    'field_type')

SENSOR_CODES = (
    'altimeter',
    'sar',
    'infrared radiometer',
    'microwave radiometer')

PLATFORM_CODES = (
    'leo satellite',
    'geostationary satellite',
    'moored buoy',
    'drifting buoy',
    'ship',
    'argo')

VARIABLE_MANDATORY_ATTRIBUTES = (
    'name',
    '_FillValue',
    'scale_factor',
    'add_offset')

VARIABLE_CF_ATTRIBUTES = (
    'units',
    'long_name',
    'valid_min',
    'valid_max',
    'standard_name')

VARIABLE_OPTIONAL_ATTRIBUTES = (
    'comment',
    'axis',
    'positive',
    'coordinates',
    'flag_meanings',
    'flag_values',
    'flag_masks',
    'depth',
    'height')

NUMERICAL_ATTRIBUTES = (
    'idf_subsampling_factor',
    'idf_spatial_resolution',
    'geospatial_lat_max',
    'geospatial_lat_min',
    'geospatial_lon_max',
    'geospatial_lon_min',
    'file_quality_level',
    'sea_floor_depth_below_sea_level',
    'site_elevation',
    'geospatial_vertical_min',
    'geospatial_vertical_max',
    'nominal_latitude',
    'nominal_longitude')

TEXT_ATTRIBUTES = (
    'idf_version',
    'idf_granule_id',
    'idf_spatial_resolution_units',
    'time_coverage_start',
    'time_coverage_end',
    'title',
    'institution',
    'comment',
    'id',
    'product_version',
    'creator_email',
    'Conventions',
    'summary',
    'references',
    'institution_abbreviation',
    'history',
    'naming_authority',
    'processing_software',
    'uuid',
    'date_created',
    'date_modified',
    'keywords',
    'keywords_vocabulary',
    'publisher_name',
    'publisher_url',
    'license',
    'netcdf_version_id',
    'spatial_resolution',
    'time_coverage_resolution',
    'geospatial_vertical_units',
    'geospatial_vertical_positive',
    'source',
    'source_version',
    'wmo_id',
    'buoy_network',
    'station_name',
    'platform_type',
    'band',
    'sensor_description',
    'sensor_manufacturer',
    'sensor_serial_number',
    'sensor_install_date',
    'sensor_height',
    'sensor_sampling_period',
    'sensor_sampling_rate',
    'sensor_calibration_date',
    'sensor_history',
    'sensor_type',
    'Metadata_Conventions',
    'metadata_link',
    'standard_name_vocabulary',
    'acknowledgement',
    'creator_name',
    'creator_url',
    'publisher_email',
    'processing_level',
    'cdm_data_type',
    'platform',
    'sensor',
    'station_id',
    'forecast_range',
    'run_time',
    'forecast_type',
    'field_type')

VARIABLE_NUMERICAL_ATTRIBUTES = (
    'valid_min',
    'valid_max')

VARIABLE_TEXT_ATTRIBUTES = (
    'name',
    'units',
    'long_name',
    'standard_name',
    'comment',
    'axis',
    'positive',
    'coordinates',
    'flag_meanings',
    'depth',
    'height')



# vim: ts=4:sts=4:sw=4:et
#
# @date 2019-09-27
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import typing
import netCDF4
import logging
import datetime
import numpy.typing
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.lib.types import Granule, TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'GRID_LATLON'

CONTROL_BITS = {'IGNORE': 0,
                'RANGE': 1,
                'SIGMA': 2,
                'CHI2': 3,
                'CHI2_P': 4,
                'CONTAMINATED': 5,
                'SUNGLINT': 6,
                'MOONGLINT': 7,
                'GAL_NOISE': 8,
                'MIXED_SCENE': 9,
                'REACH_MAXITER': 10,
                'NUM_MEAS_MIN': 11,
                'NUM_MEAS_LOW': 12,
                'MANY_OUTLIERS': 13,
                'MARQ': 14,
                'ROUGHNESS': 15,
                'FOAM': 16,
                'ECMWF': 17,
                'VALID': 18,
                'NO_SURFACE': 19,
                'RANGE_ACARD': 20,
                'SIGMA_ACARD': 21,
                'SPARE': 22,
                'USED_FARA_TEC': 23,
                'POOR_GEOPHYSICAL': 24,
                'POOR_RETRIEVAL': 25,
                'SUSPECT_RFI': 26,
                'RFI_PRONE_X': 27,
                'RFI_PRONE_Y': 28,
                'ADJUSTED_RA': 29,
                'RETRIEV_FAIL': 30,
                'SPARE2': 31}

SCIENCE_BITS = {'LAND_SEA_COAST1': 0,
                'LAND_SEA_COAST2': 1,
                'TEC_GRADIENT': 2,
                'IN_CLIM_ICE': 3,
                'ICE': 4,
                'SUSPECT_ICE': 5,
                'RAIN': 6,
                'HIGH_WIND': 7,
                'LOW_WIND': 8,
                'HIGH_SST': 9,
                'LOW_SST': 10,
                'HIGH_SSS': 11,
                'LOW_SSS': 12,
                'SEA_STATE_1': 13,
                'SEA_STATE_2': 14,
                'SEA_STATE_3': 15,
                'SEA_STATE_4': 16,
                'SEA_STATE_5': 17,
                'SEA_STATE_6': 18,
                'SST_FRONT': 19,
                'SSS_FRONT': 20,
                'ICE_ACARD': 21,
                'ECMWF_LAND': 22}

GridIndices = typing.Tuple[numpy.typing.NDArray, numpy.typing.NDArray,
                           numpy.typing.NDArray]

# Scale factor used to circumvent underflow errors
UFLOW_FIX = 10**6


class InputPathMissing(IOError):
    """"""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def get_grid_indices(lat: numpy.typing.NDArray,
                     lon: numpy.typing.NDArray,
                     x_res: float,
                     y_res: float
                     ) -> GridIndices:
    """Implement BEC method to regrid data.
    See:
    https://www.researchgate.net/publication/
    263271245_Remapping_ISEA_grid_to_lat-lon_grid_in_SMOS_context"""
    # Internal computations performed by numpy can trigger underflow errors
    # while performing divisions that invlve small numbers, so multiply both
    # the numerator and the denominator by a large number prior to computing
    # the division
    lat_rad = numpy.deg2rad(lat)
    delta = 0.5 + 1.0 * UFLOW_FIX / (2 * numpy.cos(lat_rad) * UFLOW_FIX)
    rel_i_min = -1 * delta + ((lon + 180.0) * UFLOW_FIX) / (x_res * UFLOW_FIX)
    rel_i_max = +1 * delta + ((lon + 180.0) * UFLOW_FIX) / (x_res * UFLOW_FIX)
    i_min = numpy.ceil(rel_i_min).astype('int')
    i_max = numpy.floor(rel_i_max).astype('int')
    rel_j = ((lat + 90.0) * UFLOW_FIX) / (y_res * UFLOW_FIX)
    j = numpy.floor(rel_j).astype('int')
    return (i_min, i_max, j)


def get_valid_mask(f_handler: netCDF4.Dataset,
                   lat_values: numpy.typing.NDArray,
                   sss_values: numpy.typing.NDArray
                   ) -> numpy.typing.NDArray:
    """"""
    dg_af_fov = f_handler.variables['Dg_af_fov'][:]
    control_flags = f_handler.variables['Control_Flags_corr'][:]
    science_flags = f_handler.variables['Science_Flags_corr'][:]

    control_on = numpy.int32(0)
    control_on = control_on + numpy.int32(1 << CONTROL_BITS['ECMWF'])

    control_off = numpy.int32(0)
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['CHI2_P'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['SUNGLINT'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['MOONGLINT'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['REACH_MAXITER'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['NUM_MEAS_MIN'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['NUM_MEAS_LOW'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['MANY_OUTLIERS'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['MARQ'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['SUSPECT_RFI'])

    ctrl_on_mask = (numpy.bitwise_and(control_on, control_flags) == control_on)
    ctrl_off_mask = (numpy.bitwise_and(control_off, control_flags) == 0)
    control_valid = (ctrl_on_mask & ctrl_off_mask)

    science_on = numpy.int32(0)
    science_on = science_on + numpy.int32(1 << SCIENCE_BITS['LAND_SEA_COAST1'])
    science_on = science_on + numpy.int32(1 << SCIENCE_BITS['LOW_WIND'])

    science_off = numpy.int32(0)
    science_off = science_off + numpy.int32(1 << SCIENCE_BITS['ICE'])
    science_off = science_off + numpy.int32(1 << SCIENCE_BITS['SUSPECT_ICE'])

    sci_on_mask = (numpy.bitwise_and(science_on, science_flags) == science_on)
    sci_off_mask = (numpy.bitwise_and(science_off, science_flags) == 0)
    science_valid = (sci_on_mask & sci_off_mask)

    af_fov_valid = (dg_af_fov > 130.0)

    sss_invalid_mask = (sss_values == -999)
    sss_below_25_mask: numpy.typing.NDArray[numpy.bool_] = (sss_values < 25.0)
    sss_below_30_mask: numpy.typing.NDArray[numpy.bool_] = (sss_values < 30.0)
    sss_above_40_mask: numpy.typing.NDArray[numpy.bool_] = (sss_values > 40.0)
    sss_lowlat_invalid_mask = ((numpy.abs(lat_values) <= 30.0) &
                               (sss_below_25_mask | sss_above_40_mask))
    sss_high_lat_invalid_mask = ((numpy.abs(lat_values) > 30.0) &
                                 (sss_below_30_mask | sss_above_40_mask))
    sss_invalid = (sss_invalid_mask | sss_lowlat_invalid_mask |
                   sss_high_lat_invalid_mask)

    valid_mask: numpy.typing.NDArray = (~sss_invalid &
                                        af_fov_valid &
                                        control_valid &
                                        science_valid)

    return valid_mask


def regrid_data(nlat: int,
                nlon: int,
                lat1d: numpy.typing.NDArray,
                lon1d: numpy.typing.NDArray,
                values1d: typing.Dict[str, numpy.typing.NDArray],
                valid_mask: numpy.typing.NDArray,
                granule: Granule
                ) -> typing.Tuple[numpy.typing.NDArray, numpy.typing.NDArray]:
    x_res = 360.0 / nlon
    y_res = 180.0 / nlat
    valid_ind = numpy.where(valid_mask == True)  # noqa
    valid_lat1d = lat1d[valid_ind]
    valid_lon1d = lon1d[valid_ind]
    regridded_points = get_grid_indices(valid_lat1d, valid_lon1d, x_res, y_res)

    i_lowers, i_uppers, js = regridded_points
    for var_name in values1d.keys():
        valid_var_values = values1d[var_name][valid_ind]
        _var_grid = numpy.full((nlat, nlon), 0.0, dtype=float)
        var_grid = numpy.ma.array(_var_grid)
        count_grid = numpy.full((nlat, nlon), 0, dtype=numpy.int32)
        for k, j in enumerate(js):
            i_min = i_lowers[k]
            i_max = i_uppers[k]
            if i_min >= i_max:
                i = numpy.mod(i_min, nlon).astype('int')
                count_grid[j, i] += 1
                var_grid[j, i] += valid_var_values[k]
            else:
                for _i in range(i_min, i_max):
                    i = numpy.mod(_i, nlon).astype('int')
                    count_grid[j, i] += 1
                    var_grid[j, i] += valid_var_values[k]

        # Compute mean for each cell
        positive_count_ind = numpy.where(count_grid > 0)
        var_sum = var_grid[positive_count_ind]
        count = count_grid[positive_count_ind]
        var_grid[positive_count_ind] = var_sum / count
        var_grid.mask = (count_grid <= 0)

        granule.vars[var_name]['array'] = var_grid

    # Locate surrounding rows and columns that contain no valid data in order
    # to compute the actual bounding box
    mask: numpy.typing.NDArray[numpy.bool_] = (count_grid <= 0)
    cols_mask = numpy.logical_and.reduce(mask, axis=0)
    rows_mask = numpy.logical_and.reduce(mask, axis=1)
    valid_cols_ind = numpy.where(~cols_mask)
    valid_rows_ind = numpy.where(~rows_mask)
    if 0 >= len(valid_rows_ind[0]):
        logger.warning('No valid data to process')
        raise idf_converter.lib.EarlyExit()

    lat_slice = slice(valid_rows_ind[0][0], 1 + valid_rows_ind[0][-1])
    lon_slice = slice(valid_cols_ind[0][0], 1 + valid_cols_ind[0][-1])
    lat0 = -90.0 + lat_slice.start * y_res
    lon0 = -180.0 + lon_slice.start * x_res

    # Geolocation is expected to be located at the pixel center by default
    actual_nlat = lat_slice.stop - lat_slice.start
    actual_nlon = lon_slice.stop - lon_slice.start
    lat = numpy.array([lat0 + (i + 0.5) * y_res for i in range(actual_nlat)])
    lon = numpy.array([lon0 + (i + 0.5) * x_res for i in range(actual_nlon)])

    # Crop data out of the bounding box to reduce the size of the output file.
    for var_name in granule.vars.keys():
        var_data = granule.vars[var_name]['array']
        granule.vars[var_name]['array'] = var_data[(lat_slice, lon_slice)]

    # print(lat.size, lon.size, numpy.shape(granule.vars['SSS_corr']['array']))
    return lat, lon


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """

    variables = {'SSS_corr': {'name': 'SSS_corr',
                              'valid_min': 25.0,
                              'valid_max': 40.0,
                              'options': {}},
                 'SSS_climatology': {'name': 'SSS_climatology',
                                     'valid_min': 0.0,
                                     'valid_max': 40.0,
                                     'options': {}},
                 'WS': {'name': 'WS',
                        'valid_min': 0.0,
                        'valid_max': 50.0,
                        'options': {}},
                 'SST': {'name': 'SST',
                         'valid_min': -2.10,
                         'valid_max': 36.0,
                         'options': {}}}

    channels = list(variables.keys())

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    _nlat = input_opts.get('grid_height', None)
    nlat = int(_nlat)

    _nlon = input_opts.get('grid_width', None)
    nlon = int(_nlon)

    values1d = {}
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon1d = idf_converter.lib.extract_variable_values(f_handler, 'Longitude')
    lat1d = idf_converter.lib.extract_variable_values(f_handler, 'Latitude')
    for var_id in channels:
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        _band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        values1d[var_id] = _band

    valid_mask = get_valid_mask(f_handler, lat1d, values1d['SSS_corr'])
    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)
    start_attr_name = 'Fixed_Header:Validity_Period:Validity_Start'
    start_str = f_handler.getncattr(start_attr_name)
    stop_attr_name = 'Fixed_Header:Validity_Period:Validity_Stop'
    stop_str = f_handler.getncattr(stop_attr_name)
    f_handler.close()

    lat, lon = regrid_data(nlat, nlon, lat1d, lon1d, values1d, valid_mask,
                           granule)

    granule.dims['lat'] = lat.size
    granule.dims['lon'] = lon.size
    dlon = 360.0 / nlon
    dlat = 180.0 / nlat
    spatial_resolution = min([abs(dlat), abs(dlon)]) * 111000.

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    start_dt = datetime.datetime.strptime(start_str[4:], '%Y-%m-%dT%H:%M:%S')
    stop_dt = datetime.datetime.strptime(stop_str[4:], '%Y-%m-%dT%H:%M:%S')

    granule_file = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_file)

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = spatial_resolution
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt
    granule.meta['institution'] = 'DPGS'
    granule.meta['platform'] = 'SMOS'
    granule.meta['platform_type'] = 'leo satellite'
    granule.meta['sensor'] = 'MIRAS'

    # Add transforms
    transforms: TransformsList = []

    output_opts['__export'] = channels

    yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module contains formatter methods for operations related to packing
numerical values.
"""


import typing
import logging

import numpy
import netCDF4

from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)


class ParametersRangeMissing(ValueError):
    """Error raised when a variable is missing either the valid_min attribute,
    the valid_max attribute or both.
    """
    def __init__(self, var_name: str) -> None:
        """
        Parameters
        ----------
        var_name: str
            The identifier of the variable which is missing a mandatory
            attribute
        """
        self.var_name = var_name
        msg = (f'Either the "valid_min", "valid_max" or both attributes are '
               f'not defined for variable "{var_name}".')
        super(ParametersRangeMissing, self).__init__(msg)


def as_ubytes(input_opts: InputOptions,
              output_opts: OutputOptions,
              granule: Granule,
              *args: typing.Iterable[typing.Any],
              targets: typing.Iterable[str]) -> FormatterResult:
    """
    Pack numerical values of targetted variables into ubytes and update the
    variables attributes accordingly.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    targets: tuple
        Identifers for variables whose values must be packed as ubytes

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the targetted variables
          have been packed as ubytes
    """
    for var_name in targets:
        has_valid_min = ('valid_min' in granule.vars[var_name].keys())
        has_valid_max = ('valid_max' in granule.vars[var_name].keys())

        if False in (has_valid_min, has_valid_max):
            raise ParametersRangeMissing(var_name)

        values = granule.vars[var_name]['array']
        vmin = granule.vars[var_name]['valid_min']
        vmax = granule.vars[var_name]['valid_max']
        nan_mask = ((numpy.ma.getmaskarray(values)) | (numpy.isnan(values)))

        logger.debug(f'{var_name}: [{vmin}, {vmax}]')

        # Avoid warnings caused by vmin and vmax begin masked values, which
        # may happen when a variable only contains masked values.
        if numpy.ma.is_masked(vmin):
            vmin = 0
        if numpy.ma.is_masked(vmax):
            vmax = 0

        # Converting vmin and vmax to float during scale computation avoids
        # overflow errors
        offset, scale = vmin, (float(vmax) - float(vmin)) / 254.0
        if vmin == vmax:
            scale = 1.0

        unpacked_type = values.dtype.type
        vmin = unpacked_type(vmin)
        vmax = unpacked_type(vmax)
        numpy.clip(values, vmin, vmax, out=values)

        # Required to avoid runtime warnings on masked arrays wherein the
        # division of the _FillValue by the scale cannot be stored by the dtype
        # of the array
        if isinstance(values, numpy.ma.MaskedArray):
            mask = numpy.ma.getmaskarray(values).copy()
            values[numpy.where(mask)] = vmin
            _values = (numpy.ma.getdata(values) - offset) / scale
            values.mask = mask  # Restore mask to avoid side-effects
        else:
            _values = (values - offset) / scale

        # In case the array still contains Nan values, replace them by vmin
        # before calling numpy.round to avoid errors related to type casting
        # (the rounded values corresponding to the NaN will be replaced by 255
        # anyway).
        _values[numpy.where(nan_mask)] = vmin

        result = numpy.round(_values).astype('ubyte')
        result[numpy.where(nan_mask)] = 255

        granule.vars[var_name]['array'] = result
        granule.vars[var_name]['valid_min'] = 0
        granule.vars[var_name]['valid_max'] = 254
        granule.vars[var_name]['scale_factor'] = scale
        granule.vars[var_name]['add_offset'] = offset
        granule.vars[var_name]['_FillValue'] = numpy.ubyte(255)
        granule.vars[var_name]['datatype'] = numpy.ubyte

    return input_opts, output_opts, granule


def as_float32(input_opts: InputOptions,
               output_opts: OutputOptions,
               granule: Granule,
               *args: typing.Iterable[typing.Any],
               targets: typing.Iterable[str]) -> FormatterResult:
    """"""
    fill_value = netCDF4.default_fillvals['f4']
    for var_name in targets:
        has_valid_min = ('valid_min' in granule.vars[var_name].keys())
        has_valid_max = ('valid_max' in granule.vars[var_name].keys())

        if False in (has_valid_min, has_valid_max):
            raise ParametersRangeMissing(var_name)

        values = granule.vars[var_name]['array']
        vmin = granule.vars[var_name]['valid_min']
        vmax = granule.vars[var_name]['valid_max']
        nan_mask = ((numpy.ma.getmaskarray(values)) | (numpy.isnan(values)))

        logger.debug(f'{var_name}: [{vmin}, {vmax}]')

        # Avoid warnings caused by vmin and vmax begin masked values, which
        # may happen when a variable only contains masked values.
        if numpy.ma.is_masked(vmin):
            vmin = 0
        if numpy.ma.is_masked(vmax):
            vmax = 0

        numpy.clip(values, vmin, vmax, out=values)

        result = numpy.ma.getdata(values).astype('float32')
        result[numpy.where(nan_mask)] = fill_value

        granule.vars[var_name]['array'] = result
        granule.vars[var_name]['valid_min'] = numpy.float32(vmin)
        granule.vars[var_name]['valid_max'] = numpy.float32(vmax)
        granule.vars[var_name]['_FillValue'] = fill_value
        granule.vars[var_name]['datatype'] = numpy.float32

    return input_opts, output_opts, granule



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module defines errors related to IDF v.x formatting."""

import typing
import logging
import idf_converter.lib.idf.v1.specs as idf_specs

logger = logging.getLogger(__name__)


class MissingGeolocationProjection(Exception):
    """"""
    def __init__(self) -> None:
        """"""
        msg = ('The geolocation dictionnary must have a value associated with '
               'the "projection" key.')
        super(MissingGeolocationProjection, self).__init__(msg)


class MissingGeolocationDetails(Exception):
    """"""
    def __init__(self) -> None:
        """"""
        msg = ('The geolocation dictionnary must contain either a "gcps" or a '
               '"geotransform" value.')
        super(MissingGeolocationDetails, self).__init__(msg)


class MissingGCPResolution(Exception):
    """"""
    def __init__(self) -> None:
        """"""
        msg = ('You must provide either the  number of GCPs or pixel distance '
               'between GCPs.')
        super(MissingGCPResolution, self).__init__(msg)


class DataModelNotSupported(Exception):
    """"""
    def __init__(self, data_model: str) -> None:
        """"""
        supported_models = idf_specs.SUPPORTED_MODELS.keys()
        msg = ('"{}" not found in the list of supported data models: '
               '{}.'.format(data_model, ', '.join(supported_models)))
        super(DataModelNotSupported, self).__init__(msg)


class MissingData(Exception):
    """"""
    def __init__(self) -> None:
        """"""
        msg = 'The "band" array must contain at least one element'
        super(MissingData, self).__init__(msg)


class MissingMetadata(Exception):
    """"""
    def __init__(self, attribute_names: typing.List[str]) -> None:
        """"""
        msg = ('The following metadata are required but were not provided: '
               '{}.'.format(', '.join(attribute_names)))
        super(MissingMetadata, self).__init__(msg)


class UnknownAttributes(Exception):
    """"""
    def __init__(self, attributes: typing.List[str]) -> None:
        """"""
        extra_str = ', '.join(attributes)
        msg = ('The following attributes are not supported: {}.'
               'If you really want to add them to the global attributes of '
               'the IDF file, please add them to the '
               '"allowed_extra_attributes" parameter'.format(extra_str))
        super(UnknownAttributes, self).__init__(msg)


class MissingVariableMetadata(Exception):
    """"""
    def __init__(self, attributes: typing.Dict[str, typing.List[str]]) -> None:
        """"""
        _msg = ['Some mandatory attributes are missing from the variables '
                'metadata:']
        for band in attributes:
            _msg.append('\t{}: {}'.format(band, ', '.join(attributes[band])))
        msg = '\n'.join(_msg)
        super(MissingVariableMetadata, self).__init__(msg)


class UnknownVariableAttributes(Exception):
    """"""
    def __init__(self, attributes: typing.Dict[str, typing.List[str]]) -> None:
        """"""
        _msg = ['The following variable attributes are not supported:']
        for band in attributes:
            _msg.append('\t{}: {}'.format(band, ', '.join(attributes[band])))
        _msg.append('If you really want them to be included in the IDF file, '
                    'please add them to the allowed_extra_variable_attributes '
                    'parameter.')
        msg = '\n'.join(_msg)
        super(UnknownVariableAttributes, self).__init__(msg)


class InvalidDatetimeType(Exception):
    """"""
    def __init__(self, attr: str) -> None:
        """"""
        msg = ('The "{}" attribute must be a valid datetime.datetime '
               'object.'.format(attr))
        super(InvalidDatetimeType, self).__init__(msg)


class GCPArityInconsistency(Exception):
    """"""
    def __init__(self, sizes: typing.List[int]) -> None:
        """"""
        sizes_str = ['{}'.format(s) for s in sizes]
        msg = ('The inputs provided to generate GCPs must all have the same '
               'size. The provided inputs do not respect this condition as '
               'their sizes are: {}'.format(', '.join(sizes_str)))
        super(GCPArityInconsistency, self).__init__(msg)
